# -*- coding: utf-8 -*-
"""
formlayout
==========

Module creating Qt form dialogs/layouts to edit various type of parameters


formlayout License Agreement (MIT License)
------------------------------------------

Copyright (c) 2009-2015 Pierre Raybaut

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import print_function

__version__ = '2.0.0alpha'
__license__ = __doc__

DEBUG_FORMLAYOUT = False

import os
import sys
import datetime

STDERR = sys.stderr


# ---+- PyQt-PySide compatibility -+----
_modname = os.environ.setdefault('QT_API', 'pyqt')
assert _modname in ('pyqt', 'pyqt5', 'pyside', 'pyside2')

if os.environ['QT_API'].startswith('pyqt'):
    try:
        if os.environ['QT_API'] == 'pyqt5':
            import PyQt5  # analysis:ignore
        else:
            import PyQt4  # analysis:ignore
    except ImportError:
        # Switching to PySide
        os.environ['QT_API'] = _modname = 'pyside'
        try:
            import PySide  # analysis:ignore
        except ImportError:
            # Switching to PySide2
            os.environ['QT_API'] = _modname = 'pyside2'
            try:
                import PySide2  # analysis:ignore
            except ImportError:
                raise ImportError("formlayout requires PyQt4, PyQt5, PySide or PySide2")

if os.environ['QT_API'] == 'pyqt':
    try:
        from PyQt4.QtGui import QFormLayout
    except ImportError:
        raise ImportError("formlayout requires PyQt4, PyQt5, PySide or PySide2")
    from PyQt4.QtGui import *  # analysis:ignore
    from PyQt4.QtCore import *  # analysis:ignore
    from PyQt4.QtCore import pyqtSlot as Slot
    from PyQt4.QtCore import pyqtProperty as Property
    QT_LIB = 'PyQt4'

if os.environ['QT_API'] == 'pyqt5':
    from PyQt5.QtWidgets import *  # analysis:ignore
    from PyQt5.QtGui import *  # analysis:ignore
    from PyQt5.QtCore import *  # analysis:ignore
    from PyQt5.QtCore import pyqtSignal as Signal  # analysis:ignore
    from PyQt5.QtCore import pyqtSlot as Slot  # analysis:ignore
    from PyQt5.QtCore import pyqtProperty as Property  # analysis:ignore
    SIGNAL = None  # analysis:ignore
    QT_LIB = 'PyQt5'

if os.environ['QT_API'] == 'pyside':
    from PySide.QtGui import *  # analysis:ignore
    from PySide.QtCore import *  # analysis:ignore
    QT_LIB = 'PySide'
    
if os.environ['QT_API'] == 'pyside2':
    from PySide2.QtGui import *  # analysis:ignore
    from PySide2.QtCore import *  # analysis:ignore
    from PySide2.QtWidgets import *  # analysis:ignore
    QT_LIB = 'PySide2'


# ---+- Python 2-3 compatibility -+----
PY2 = sys.version[0] == '2'

if PY2:
    # Python 2
    import codecs
    def u(obj):
        """Make unicode object"""
        return codecs.unicode_escape_decode(obj)[0]
else:
    # Python 3
    def u(obj):
        """Return string as it is"""
        return obj

def is_text_string(obj):
    """Return True if `obj` is a text string, False if it is anything else,
    like binary data (Python 3) or QString (Python 2, PyQt API #1)"""
    if PY2:
        # Python 2
        return isinstance(obj, basestring)
    else:
        # Python 3
        return isinstance(obj, str)

def is_binary_string(obj):
    """Return True if `obj` is a binary string, False if it is anything else"""
    if PY2:
        # Python 2
        return isinstance(obj, str)
    else:
        # Python 3
        return isinstance(obj, bytes)

def is_string(obj):
    """Return True if `obj` is a text or binary Python string object,
    False if it is anything else, like a QString (Python 2, PyQt API #1)"""
    return is_text_string(obj) or is_binary_string(obj)

def to_text_string(obj, encoding=None):
    """Convert `obj` to (unicode) text string"""
    if PY2:
        # Python 2
        if encoding is None:
            return unicode(obj)
        else:
            return unicode(obj, encoding)
    else:
        # Python 3
        if encoding is None:
            return str(obj)
        elif isinstance(obj, str):
            # In case this function is not used properly, this could happen
            return obj
        else:
            return str(obj, encoding)


class ColorButton(QPushButton):
    """
    Color choosing push button
    """
    __pyqtSignals__ = ("colorChanged(QColor)",)
    if SIGNAL is None:
        colorChanged = Signal("QColor")
    
    def __init__(self, parent=None, colorChanged_slot=None):
        QPushButton.__init__(self, parent)
        self.setFixedSize(20, 20)
        self.setIconSize(QSize(12, 12))
        
        if colorChanged_slot:
            if SIGNAL is None:
                self.colorChanged.connect(colorChanged_slot)
            else:
                self.connect(self, SIGNAL("colorChanged(QColor)"),
                             colorChanged_slot)
        
        if SIGNAL is None:
            self.clicked.connect(self.choose_color)
        else:
            self.connect(self, SIGNAL("clicked()"), self.choose_color)
        self._color = QColor()
    
    def choose_color(self):
        color = QColorDialog.getColor(self._color, self.parentWidget())
        if color.isValid():
            self.set_color(color)
    
    def get_color(self):
        return self._color
    
    @Slot(QColor)
    def set_color(self, color):
        if color != self._color:
            self._color = color
            if SIGNAL is None:
                self.colorChanged.emit(self._color)
            else:
                self.emit(SIGNAL("colorChanged(QColor)"), self._color)
            pixmap = QPixmap(self.iconSize())
            pixmap.fill(color)
            self.setIcon(QIcon(pixmap))
    
    color = Property("QColor", get_color, set_color)


def text_to_qcolor(text):
    """
    Create a QColor from specified string
    Avoid warning from Qt when an invalid QColor is instantiated
    """
    color = QColor()
    if not is_string(text): # testing for QString (PyQt API#1)
        text = str(text)
    if not is_text_string(text):
        return color
    if text.startswith('#') and len(text)==7:
        correct = '#0123456789abcdef'
        for char in text:
            if char.lower() not in correct:
                return color
    elif text not in list(QColor.colorNames()):
        return color
    color.setNamedColor(text)
    return color


class ColorLayout(QHBoxLayout):
    """Color-specialized QLineEdit layout"""
    def __init__(self, color, parent=None):
        QHBoxLayout.__init__(self)
        assert isinstance(color, QColor)
        self.lineedit = QLineEdit(color.name(), parent)
        if SIGNAL is None:
            self.lineedit.textChanged.connect(self.update_color)
        else:
            self.connect(self.lineedit, SIGNAL("textChanged(QString)"),
                         self.update_color)
        self.addWidget(self.lineedit)
        self.colorbtn = ColorButton(parent, self.update_text)
        self.colorbtn.color = color
        
        self.addWidget(self.colorbtn)

    def update_color(self, text):
        color = text_to_qcolor(text)
        if color.isValid():
            self.colorbtn.color = color

    def update_text(self, color):
        self.lineedit.setText(color.name())
        
    def text(self):
        return self.lineedit.text()

    def setStyleSheet(self, style):
        self.lineedit.setStyleSheet(style)
        self.colorbtn.setStyleSheet(style)


class FileLayout(QHBoxLayout):
    """File-specialized QLineEdit layout"""
    def __init__(self, value, parent=None):
        QHBoxLayout.__init__(self)
        self.value = value
        self.lineedit = QLineEdit('', parent)
        self.addWidget(self.lineedit)
        self.filebtn = QPushButton('Browse')
        self.filebtn.clicked.connect(self.getfile)
        self.addWidget(self.filebtn)

    def getfile(self):
        if self.value.startswith('file'):
            name = QFileDialog.getOpenFileName(None, 'Select file',
                                               filter=self.value[5:])
            if QT_LIB == 'PyQt5':
                name, _filter = name
        elif self.value == 'dir':
            name = QFileDialog.getExistingDirectory(None, 'Select directory')
        if name:
            self.lineedit.setText(name)

    def text(self):
        return self.lineedit.text()

    def setStyleSheet(self, style):
        self.lineedit.setStyleSheet(style)


class SliderLayout(QHBoxLayout):
    """QSlider with QLabel"""
    def __init__(self, value, parent=None):
        QHBoxLayout.__init__(self)
        index = value.find('@')
        if index != -1:
            value, default = value[:index], int(value[index+1:])
        else:
            default = False
        parsed = value.split(':')
        self.slider = QSlider(Qt.Horizontal)
        if parsed[-1] == '':
            self.slider.setTickPosition(2)
            parsed.pop(-1)
        if len(parsed) == 2:
            self.slider.setMaximum(int(parsed[1]))
        elif len(parsed) == 3:
            self.slider.setMinimum(int(parsed[1]))
            self.slider.setMaximum(int(parsed[2]))
        if default:
            self.slider.setValue(default)  # always set value in last
        if SIGNAL is None:
            self.slider.valueChanged.connect(self.update)
        else:
            self.connect(self.slider, SIGNAL("valueChanged(int)"), self.update)
        self.cpt = QLabel(str(self.value()))
        self.addWidget(self.slider)
        self.addWidget(self.cpt)

    def update(self):
        self.cpt.setText(str(self.value()))

    def value(self):
        return self.slider.value()

    def setStyleSheet(self, style):
        self.slider.setStyleSheet(style)
        self.cpt.setStyleSheet(style)


class RadioLayout(QVBoxLayout):
    """Radio buttons layout with QButtonGroup"""
    def __init__(self, buttons, index, parent=None):
        QVBoxLayout.__init__(self)
        self.setSpacing(0)
        self.group = QButtonGroup()
        for i, button in enumerate(buttons):
            btn = QRadioButton(button)
            if i == index:
                btn.setChecked(True)
            self.addWidget(btn)
            self.group.addButton(btn, i)

    def currentIndex(self):
        return self.group.checkedId()

    def setStyleSheet(self, style):
        for btn in self.group.buttons():
            btn.setStyleSheet(style)


class CheckLayout(QVBoxLayout):
    """Check boxes layout with QButtonGroup"""
    def __init__(self, boxes, checks, parent=None):
        QVBoxLayout.__init__(self)
        self.setSpacing(0)
        self.group = QButtonGroup()
        self.group.setExclusive(False)
        for i, (box, check) in enumerate(zip(boxes, checks)):
            cbx = QCheckBox(box)
            cbx.setChecked(eval(check))
            self.addWidget(cbx)
            self.group.addButton(cbx, i)

    def values(self):
        return [cbx.isChecked() for cbx in self.group.buttons()]

    def setStyleSheet(self, style):
        for cbx in self.group.buttons():
            cbx.setStyleSheet(style)


class PushLayout(QHBoxLayout):
    """Push buttons horizontal layout"""
    def __init__(self, buttons, parent=None):
        QHBoxLayout.__init__(self)
        self.result = parent.result
        self.dialog = parent.get_dialog()
        for button in buttons:
            label, callback = button
            self.btn = QPushButton(label)
            if SIGNAL is None:
                self.btn.clicked.connect(self.call(callback))
            else:
                self.connect(self.btn, SIGNAL("clicked()"), self.call(callback))
            self.addWidget(self.btn)

    def call(self, callback):
        return lambda: self.apply(callback)

    def apply(self, callback):
        if self.result == 'XML':
            app = ET.Element('App')
            app.attrib['title'] = self.dialog.title
            child = ET.fromstring(self.dialog.formwidget.get())
            app.append(child)
            callback(ET.tostring(app),
                     self.dialog.formwidget.get_widgets())
        else:
            callback(self.dialog.formwidget.get(),
                     self.dialog.formwidget.get_widgets())


class CountLayout(QHBoxLayout):
    """Field with a QSpinBox"""
    def __init__(self, field, parent=None):
        QHBoxLayout.__init__(self)
        self.field = field
        self.count = QSpinBox()
        self.count.setFixedWidth(45)
        self.addWidget(self.field)
        self.addWidget(self.count)

    def text(self):
        return self.field.text()

    def currentIndex(self):
        return self.field.currentIndex()

    def n(self):
        return self.count.value()

    def setStyleSheet(self, style):
        self.field.setStyleSheet(style)
        self.count.setStyleSheet(style)


def font_is_installed(font):
    """Check if font is installed"""
    return [fam for fam in QFontDatabase().families()
            if to_text_string(fam) == font]

def tuple_to_qfont(tup):
    """
    Create a QFont from tuple:
        (family [string], size [int], italic [bool], bold [bool])
    """
    if not isinstance(tup, tuple) or len(tup) != 4 \
       or not is_text_string(tup[0]) \
       or not isinstance(tup[1], int) \
       or not isinstance(tup[2], bool) \
       or not isinstance(tup[3], bool):
        return None
    font = QFont()
    family, size, italic, bold = tup
    font.setFamily(family)
    font.setPointSize(size)
    font.setItalic(italic)
    font.setBold(bold)
    return font

def qfont_to_tuple(font):
    return (to_text_string(font.family()), int(font.pointSize()),
            font.italic(), font.bold())

class FontLayout(QGridLayout):
    """Font selection"""
    def __init__(self, value, parent=None):
        QGridLayout.__init__(self)
        if not font_is_installed(value[0]):
            print("Warning: Font `%s` is not installed" % value[0],
                  file=sys.stderr)
        font = tuple_to_qfont(value)
        assert font is not None
        
        # Font family
        self.family = QFontComboBox(parent)
        self.family.setCurrentFont(font)
        self.addWidget(self.family, 0, 0, 1, -1)
        
        # Font size
        self.size = QComboBox(parent)
        self.size.setEditable(True)
        sizelist = list(range(6, 12)) + list(range(12, 30, 2)) + [36, 48, 72]
        size = font.pointSize()
        if size not in sizelist:
            sizelist.append(size)
            sizelist.sort()
        self.size.addItems([str(s) for s in sizelist])
        self.size.setCurrentIndex(sizelist.index(size))
        self.addWidget(self.size, 1, 0)
        
        # Italic or not
        self.italic = QCheckBox(self.tr("Italic"), parent)
        self.italic.setChecked(font.italic())
        self.addWidget(self.italic, 1, 1)
        
        # Bold or not
        self.bold = QCheckBox(self.tr("Bold"), parent)
        self.bold.setChecked(font.bold())
        self.addWidget(self.bold, 1, 2)
        
    def get_font(self):
        font = self.family.currentFont()
        font.setItalic(self.italic.isChecked())
        font.setBold(self.bold.isChecked())
        font.setPointSize(int(self.size.currentText()))
        return qfont_to_tuple(font)

    def setStyleSheet(self, style):
        self.family.setStyleSheet(style)
        self.size.setStyleSheet(style)
        self.italic.setStyleSheet(style)
        self.bold.setStyleSheet(style)


def is_float_valid(edit):
    text = edit.text()
    state = edit.validator().validate(text, 0)[0]
    return state == QDoubleValidator.Acceptable

def is_required_valid(edit, widget_color):
    required_color = "background-color:rgb(255, 175, 90);"
    if widget_color:
        widget_color = "background-color:" + widget_color + ";"
    else:
        widget_color = ""
    if isinstance(edit, (QLineEdit, FileLayout)):
        if edit.text():
            edit.setStyleSheet(widget_color)
            return True
        else:
            edit.setStyleSheet(required_color)
    elif isinstance(edit, (QComboBox, RadioLayout)):
        if edit.currentIndex() != -1:
            edit.setStyleSheet(widget_color)
            return True
        else:
            edit.setStyleSheet(required_color)
    elif isinstance(edit, QTextEdit):
        if edit.toPlainText():
            edit.setStyleSheet(widget_color)
            return True
        else:
            edit.setStyleSheet(required_color)
    return False

class FormWidget(QWidget):
    def __init__(self, data, comment="", parent=None):
        QWidget.__init__(self, parent)
        from copy import deepcopy
        self.data = deepcopy(data)
        self.result = parent.result
        self.type = parent.type
        self.widget_color = parent.widget_color
        self.widgets = []
        self.formlayout = QFormLayout(self)
        if comment:
            self.formlayout.addRow(QLabel(comment))
            self.formlayout.addRow(QLabel(" "))
        if DEBUG_FORMLAYOUT:
            print("\n"+("*"*80))
            print("DATA:", self.data)
            print("*"*80)
            print("COMMENT:", comment)
            print("*"*80)
            
    def get_dialog(self):
        """Return FormDialog instance"""
        dialog = self.parent()
        while not isinstance(dialog, QDialog):
            dialog = dialog.parent()
        return dialog

    def setup(self):
        for label, value in self.data:
            if DEBUG_FORMLAYOUT:
                print("value:", value)
            if label is None and value is None:
                # Separator: (None, None)
                separator = QFrame()
                separator.setFrameShape(QFrame.HLine)
                separator.setFrameShadow(QFrame.Sunken)
                self.formlayout.addRow(separator)
                self.widgets.append(None)
                continue
            if label is None:
                if isinstance(value, (list, tuple)):
                    field = PushLayout(value, self)
                    self.formlayout.addRow(field)
                else:
                    img_fmt = tuple(['.'+str(bytes(ext.data() if QT_LIB == "PySide" or QT_LIB == "PySide2" else ext).decode()) 
                                    for ext in QImageReader.supportedImageFormats()])
                    if value.endswith(img_fmt):
                        # Image
                        pixmap = QPixmap(value)
                        lab = QLabel()
                        lab.setPixmap(pixmap)
                        self.formlayout.addRow(lab)
                    else:
                        # Comment
                        self.formlayout.addRow(QLabel(value))
                self.widgets.append(None)
                continue
            if tuple_to_qfont(value) is not None:
                field = FontLayout(value, self)
            elif text_to_qcolor(value).isValid():
                field = ColorLayout(QColor(value), self)
            elif is_text_string(value):
                if value in ['file', 'dir'] or value.startswith('file:'):
                    field = FileLayout(value, self)
                elif value == 'slider' or value.startswith('slider:') \
                                       or value.startswith('slider@'):
                    field = SliderLayout(value, self)
                elif value == 'password':
                    field = QLineEdit(self)
                    field.setEchoMode(QLineEdit.Password)
                elif value in ['calendar', 'calendarM'] \
                           or value.startswith(('calendar:', 'calendarM:')) \
                           or value.startswith(('calendar@', 'calendarM@')):
                    index = value.find('@')
                    if index != -1:
                        value, date = value[:index], value[index+1:]
                    else:
                        date = False
                    field = QCalendarWidget(self)
                    field.setVerticalHeaderFormat(field.NoVerticalHeader)
                    parsed = value.split(':')
                    if parsed[-1] == '':
                        field.setGridVisible(True)
                        parsed.pop(-1)
                    if parsed[0] == 'calendarM':
                        field.setFirstDayOfWeek(Qt.Monday)
                    if len(parsed) == 2:
                        field.setMaximumDate(datetime.date(*eval(parsed[1])))
                    elif len(parsed) == 3:
                        field.setMinimumDate(datetime.date(*eval(parsed[1])))
                        field.setMaximumDate(datetime.date(*eval(parsed[2])))
                    if date:
                        field.setSelectedDate(datetime.date(*eval(date)))
                elif '\n' in value:
                    if value == '\n':
                        value = ''
                    for linesep in (os.linesep, '\n'):
                        if linesep in value:
                            value = value.replace(linesep, u("\u2029"))
                    field = QTextEdit(value, self)
                else:
                    field = QLineEdit(value, self)
            elif isinstance(value, (list, tuple)) and is_text_string(value[0])\
                                                 and value[0].startswith('0b'):
                field = CheckLayout(value[1:], value[0][2:], self)
            elif isinstance(value, (list, tuple)):
                save_value = value
                value = list(value)  # always needed to protect self.data
                selindex = value.pop(0)
                if isinstance(selindex, int):
                    selindex = selindex - 1
                if isinstance(value[0], (list, tuple)):
                    keys = [ key for key, _val in value ]
                    value = [ val for _key, val in value ]
                else:
                    keys = value
                if selindex in value:
                    selindex = value.index(selindex)
                elif selindex in keys:
                    selindex = keys.index(selindex)
                elif not isinstance(selindex, int):
                    print("Warning: '%s' index is invalid (label: "\
                          "%s, value: %s)" % (selindex, label, value),
                          file=STDERR)
                    selindex = -1
                if isinstance(save_value, list):
                    field = QComboBox(self)
                    field.addItems(value)
                    field.setCurrentIndex(selindex)
                elif isinstance(save_value, tuple):
                    field = RadioLayout(value, selindex, self)
            elif isinstance(value, bool):
                field = QCheckBox(self)
                field.setChecked(value)
            elif isinstance(value, float):
                field = QLineEdit(QLocale().toString(value), self)
                field.setValidator(QDoubleValidator(field))
                dialog = self.get_dialog()
                dialog.register_float_field(field)
                if SIGNAL is None:
                    field.textChanged.connect(dialog.float_valid)
                else:
                    self.connect(field, SIGNAL('textChanged(QString)'),
                                 dialog.float_valid)
            elif isinstance(value, int):
                field = QSpinBox(self)
                field.setRange(-1e9, 1e9)
                field.setValue(value)
            elif isinstance(value, datetime.datetime):
                field = QDateTimeEdit(self)
                field.setDateTime(value)
            elif isinstance(value, datetime.date):
                field = QDateEdit(self)
                field.setDate(value)
            elif isinstance(value, datetime.time):
                field = QTimeEdit(self)
                field.setTime(value)
            else:
                field = QLineEdit(repr(value), self)

            # Eventually catching the 'countfield' feature and processing it
            if label.startswith('n '):
                label = label[2:]
                if isinstance(field, QLineEdit) and is_text_string(value) or\
                   isinstance(field, QComboBox):
                    field = CountLayout(field)
                else:
                    print("Warning: '%s' doesn't support 'nfield' feature"\
                          % label, file=STDERR)

            # Eventually extracting tooltip from label and processing it
            index = label.find('::')
            if index != -1:
                label, tooltip = label[:index], label[index+2:]
                field.setToolTip(tooltip)

            # Eventually catching the 'required' feature and processing it
            if label.endswith(' *'):
                label = label[:-1] + '<font color="red">*</font>'
                if isinstance(field, (QLineEdit, QTextEdit, QComboBox,
                                      FileLayout, RadioLayout)):
                    dialog = self.get_dialog()
                    dialog.register_required_field(field)
                else:
                    print("Warning: '%s' doesn't support 'required' feature"\
                          % type(field), file=STDERR)
                if isinstance(field, QLineEdit):
                    if SIGNAL is None:
                        field.textChanged.connect(dialog.required_valid)
                    else:
                        self.connect(field, SIGNAL('textChanged(QString)'),
                                     dialog.required_valid)
                elif isinstance(field, QTextEdit):
                    if SIGNAL is None:
                        field.textChanged.connect(dialog.required_valid)
                    else:
                        self.connect(field, SIGNAL('textChanged()'),
                                     dialog.required_valid)
                elif isinstance(field, QComboBox):
                    if SIGNAL is None:
                        field.currentIndexChanged.connect(\
                            dialog.required_valid)
                    else:
                        self.connect(field,
                                     SIGNAL('currentIndexChanged(QString)'),
                                     dialog.required_valid)
                elif isinstance(field, FileLayout):
                    if SIGNAL is None:
                        field.lineedit.textChanged.connect(\
                            dialog.required_valid)
                    else:
                        self.connect(field.lineedit,
                                     SIGNAL('textChanged(QString)'),
                                     dialog.required_valid)
                elif isinstance(field, RadioLayout):
                    if SIGNAL is None:
                        field.group.buttonClicked.connect(\
                            dialog.required_valid)
                    else:
                        self.connect(field.group, SIGNAL('buttonClicked(int)'),
                                     dialog.required_valid)

            # Eventually setting the widget_color
            if self.widget_color:
                style = "background-color:" + self.widget_color + ";"
                field.setStyleSheet(style)

            if self.type == 'form':
                self.formlayout.addRow(label, field)
            elif self.type == 'questions':
                self.formlayout.addRow(QLabel(label))
                self.formlayout.addRow(field)
            self.widgets.append(field)
            
    def get(self):
        valuelist = []
        for index, (label, value) in enumerate(self.data):
            field = self.widgets[index]
            if label is None:
                # Separator / Comment
                continue
            if label.startswith('n '):
                label = label[2:]
            if tuple_to_qfont(value) is not None:
                value = field.get_font()
            elif is_text_string(value):
                if isinstance(field, QTextEdit):
                    value = to_text_string(field.toPlainText()
                                           ).replace(u("\u2029"), os.linesep)
                elif isinstance(field, SliderLayout):
                    value = field.value()
                elif isinstance(field, QCalendarWidget):
                    value = field.selectedDate()
                    try:
                        value = value.toPyDate()  # PyQt
                    except AttributeError:
                        value = value.toPython()  # PySide
                else:
                    value = to_text_string(field.text())
            elif isinstance(field, CheckLayout):
                value = field.values()
            elif isinstance(value, (list, tuple)):
                index = int(field.currentIndex())
                if isinstance(value[0], int):
                    # Return an int index, if initialization was an int
                    value = index + 1
                else:
                    value = value[index+1]
                    if isinstance(value, (list, tuple)):
                        value = value[0]
            elif isinstance(value, bool):
                value = field.checkState() == Qt.Checked
            elif isinstance(value, float):
                value = float(QLocale().toDouble(field.text())[0])
            elif isinstance(value, int):
                value = int(field.value())
            elif isinstance(value, datetime.datetime):
                value = field.dateTime()
                try:
                    value = value.toPyDateTime()  # PyQt
                except AttributeError:
                    value = value.toPython()  # PySide
            elif isinstance(value, datetime.date):
                value = field.date()
                try:
                    value = value.toPyDate()  # PyQt
                except AttributeError:
                    value = value.toPython()  # PySide
            elif isinstance(value, datetime.time):
                value = field.time()
                try:
                    value = value.toPyTime()  # PyQt
                except AttributeError:
                    value = value.toPython()  # PySide
            else:
                value = eval(str(field.text()))
            if isinstance(field, CountLayout):
                value = (value, field.n())
            valuelist.append((label, value))
        if self.result == 'list':
            return [value for label, value in valuelist]
        elif self.result in ['dict', 'OrderedDict', 'JSON']:
            if self.result == 'dict':
                dic = {}
            else:
                dic = OrderedDict()
            for label, value in valuelist:
                if label in dic.keys():
                    print("Warning: '%s' is duplicate and '%s' doesn't "\
                          "handle it, you should use 'list' or 'XML' instead"\
                          % (label, self.result), file=STDERR)
                if isinstance(value, (datetime.date, datetime.time,
                              datetime.datetime)) and self.result == 'JSON':
                    dic[label] = value.isoformat()
                else:
                    dic[label] = value
            if self.result == 'JSON':
                return json.dumps(dic)
            else:
                return dic
        elif self.result == 'XML':
            form = ET.Element('Form')
            for label, value in valuelist:
                tooltip = ''
                index = label.find('::')
                if index != -1:
                    label, tooltip = label[:index], label[index+2:]
                required = 'false'
                if label.endswith(' *'):
                    label = label[:-2]
                    required = 'true'
                child = ET.SubElement(form, label)
                if isinstance(value, tuple):
                    child.text = to_text_string(value[0])
                    child.attrib['amount'] = to_text_string(value[1])
                else:
                    if isinstance(value, datetime.datetime):
                        child.text = value.isoformat()
                    else:
                        child.text = to_text_string(value)
                child.attrib['tooltip'] = tooltip
                child.attrib['required'] = required
            return ET.tostring(form)

    def get_widgets(self):
        return self.widgets


class FormComboWidget(QWidget):
    def __init__(self, datalist, comment="", parent=None):
        QWidget.__init__(self, parent)
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.combobox = QComboBox()
        layout.addWidget(self.combobox)
        
        self.stackwidget = QStackedWidget(self)
        layout.addWidget(self.stackwidget)
        if SIGNAL is None:
            self.combobox.currentIndexChanged.connect(
                                            self.stackwidget.setCurrentIndex)
        else:
            self.connect(self.combobox, SIGNAL("currentIndexChanged(int)"),
                         self.stackwidget, SLOT("setCurrentIndex(int)"))

        self.result = parent.result 
        self.widget_color = parent.widget_color
        if self.widget_color:
            style = "background-color:" + self.widget_color + ";"
            self.combobox.setStyleSheet(style)
        self.type = parent.type
        self.widgetlist = []
        for data, title, comment in datalist:
            self.combobox.addItem(title)
            widget = FormWidget(data, comment=comment, parent=self)
            self.stackwidget.addWidget(widget)
            self.widgetlist.append((title, widget))
            
    def setup(self):
        for title, widget in self.widgetlist:
            widget.setup()

    def get(self):
        if self.result == 'list':
            return [widget.get() for title, widget in self.widgetlist]
        elif self.result in ['dict', 'OrderedDict', 'JSON']:
            if self.result == 'dict':
                dic = {}
            else:
                dic = OrderedDict()
            for title, widget in self.widgetlist:
                if self.result == 'JSON':
                    dic[title] = json.loads(widget.get(),
                                            object_pairs_hook=OrderedDict)
                else:
                    dic[title] = widget.get()
            if self.result == 'JSON':
                return json.dumps(dic)
            else:
                return dic
        elif self.result == 'XML':
            combos = ET.Element('Combos')
            for title, widget in self.widgetlist:
                combo = ET.SubElement(combos, 'Combo')
                combo.attrib['title'] = title
                child = ET.fromstring(widget.get())
                combo.append(child)
            return ET.tostring(combos)

    def get_widgets(self):
        widgets = []
        for title, widget in self.widgetlist:
            widgets.extend(widget.get_widgets())
        return widgets


class FormTabWidget(QWidget):
    def __init__(self, datalist, comment="", parent=None):
        QWidget.__init__(self, parent)
        layout = QVBoxLayout()
        self.tabwidget = QTabWidget()
        layout.addWidget(self.tabwidget)
        self.setLayout(layout)
        self.result = parent.result
        self.widget_color = parent.widget_color
        self.type = parent.type
        self.widgetlist = []
        for data, title, comment in datalist:
            if len(data[0])==3:
                widget = FormComboWidget(data, comment=comment, parent=self)
            else:
                widget = FormWidget(data, comment=comment, parent=self)
            index = self.tabwidget.addTab(widget, title)
            self.tabwidget.setTabToolTip(index, comment)
            self.widgetlist.append((title, widget))
            
    def setup(self):
        for title, widget in self.widgetlist:
            widget.setup()
            
    def get(self):
        if self.result == 'list':
            return [widget.get() for title, widget in self.widgetlist]
        elif self.result in ['dict', 'OrderedDict', 'JSON']:
            if self.result == 'dict':
                dic = {}
            else:
                dic = OrderedDict()
            for title, widget in self.widgetlist:
                if self.result == 'JSON':
                    dic[title] = json.loads(widget.get(),
                                            object_pairs_hook=OrderedDict)
                else:
                    dic[title] = widget.get()
            if self.result == 'JSON':
                return json.dumps(dic)
            else:
                return dic
        elif self.result == 'XML':
            tabs = ET.Element('Tabs')
            for title, widget in self.widgetlist:
                tab = ET.SubElement(tabs, 'Tab')
                tab.attrib['title'] = title
                child = ET.fromstring(widget.get())
                tab.append(child)
            return ET.tostring(tabs)

    def get_widgets(self):
        widgets = []
        for title, widget in self.widgetlist:
            widgets.extend(widget.get_widgets())
        return widgets


class FormDialog(QDialog):
    """Form Dialog"""
    def __init__(self, data, title="", comment="", icon=None, parent=None,
                 apply=None, ok=None, cancel=None, result=None, outfile=None,
                 type=None, scrollbar=None, background_color=None,
                 widget_color=None):
        QDialog.__init__(self, parent)
        
        # Destroying the C++ object right after closing the dialog box,
        # otherwise it may be garbage-collected in another QThread
        # (e.g. the editor's analysis thread in Spyder), thus leading to
        # a segmentation fault on UNIX or an application crash on Windows
        self.setAttribute(Qt.WA_DeleteOnClose)

        self.type = type
        self.title = title
        self.ok = ok
        self.cancel = cancel
        self.apply_ = None
        self.apply_callback = None
        if callable(apply):
            self.apply_callback = apply
        elif isinstance(apply, (list, tuple)):
            self.apply_, self.apply_callback = apply
        elif apply is not None:
            raise AssertionError("`apply` argument must be either a function "\
                                 "or tuple ('Apply label', apply_callback)")
        self.outfile = outfile
        self.result = result
        if self.result in ['OrderedDict', 'JSON']:
            global OrderedDict
            from collections import OrderedDict
            if self.result == 'JSON':
                global json
                import json
        elif self.result == 'XML':
            global ET
            import xml.etree.ElementTree as ET

        self.widget_color = widget_color
        if background_color:
            style = "FormDialog {background-color:" + background_color + ";}"
            self.setStyleSheet(style)

        # Form
        if isinstance(data[0][0], (list, tuple)):
            self.formwidget = FormTabWidget(data, comment=comment,
                                            parent=self)
        elif len(data[0])==3:
            self.formwidget = FormComboWidget(data, comment=comment,
                                              parent=self)
        else:
            self.formwidget = FormWidget(data, comment=comment, 
                                         parent=self)
        layout = QVBoxLayout()

        if scrollbar == True:
            scroll = QScrollArea(parent=self)
            scroll.setWidgetResizable(True)
            scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
            scroll.setWidget(self.formwidget)
            layout.addWidget(scroll)
        else:
            layout.addWidget(self.formwidget)

        self.float_fields = []
        self.required_fields = []
        self.formwidget.setup()
        
        # Button box
        self.bbox = bbox = QDialogButtonBox()
        if self.ok == True:
            bbox.addButton(QDialogButtonBox.Ok)
        elif self.ok:
            ok_btn = QPushButton(self.ok)
            bbox.addButton(ok_btn, QDialogButtonBox.AcceptRole)
        if self.cancel == True:
            bbox.addButton(QDialogButtonBox.Cancel)
        elif self.cancel:
            cancel_btn = QPushButton(self.cancel)
            bbox.addButton(cancel_btn, QDialogButtonBox.RejectRole)

        if self.apply_callback is not None:
            if self.apply_:
                apply_btn = QPushButton(self.apply_)
                bbox.addButton(apply_btn, QDialogButtonBox.ApplyRole)
            else:
                apply_btn = bbox.addButton(QDialogButtonBox.Apply)
            if SIGNAL is None:
                apply_btn.clicked.connect(self.apply)
            else:
                self.connect(apply_btn, SIGNAL("clicked()"), self.apply)
        if SIGNAL is None:
            if self.ok:
                bbox.accepted.connect(self.accept)
            if self.cancel:
                bbox.rejected.connect(self.reject)
        else:
            if self.ok:
                self.connect(bbox, SIGNAL("accepted()"), SLOT("accept()"))
            if self.cancel:
                self.connect(bbox, SIGNAL("rejected()"), SLOT("reject()"))
        layout.addWidget(bbox)
        self.required_valid()

        self.setLayout(layout)
        
        self.setWindowTitle(self.title)
        if not isinstance(icon, QIcon):
            icon = QWidget().style().standardIcon(QStyle.SP_MessageBoxQuestion)
        self.setWindowIcon(icon)
        
    def register_float_field(self, field):
        self.float_fields.append(field)

    def register_required_field(self, field):
        self.required_fields.append(field)

    def float_valid(self):
        valid = True
        for field in self.float_fields:
            if not is_float_valid(field):
                valid = False
        self.update_buttons(valid)

    def required_valid(self):
        valid = True
        for field in self.required_fields:
            if not is_required_valid(field, self.widget_color):
                valid = False
        self.update_buttons(valid)

    def update_buttons(self, valid):
        for btn in self.bbox.buttons():
            btn_role = self.bbox.buttonRole(btn)
            if btn_role in (QDialogButtonBox.AcceptRole,
                            QDialogButtonBox.ApplyRole):
                btn.setEnabled(valid)
        
    def accept(self):
        if self.result == 'XML':
            app = ET.Element('App')
            app.attrib['title'] = self.title
            child = ET.fromstring(self.formwidget.get())
            app.append(child)
            self.data = ET.tostring(app)
        else:
            self.data = self.formwidget.get()
        QDialog.accept(self)
        
    def reject(self):
        self.data = None
        QDialog.reject(self)
        
    def apply(self):
        if self.result == 'XML':
            app = ET.Element('App')
            app.attrib['title'] = self.title
            child = ET.fromstring(self.formwidget.get())
            app.append(child)
            self.apply_callback(ET.tostring(app),
                                self.formwidget.get_widgets())
        else:
            self.apply_callback(self.formwidget.get(),
                                self.formwidget.get_widgets())
        
    def get(self):
        """Return form result"""
        # It is import to avoid accessing Qt C++ object as it has probably
        # already been destroyed, due to the Qt.WA_DeleteOnClose attribute
        if self.outfile:
            if self.result in ['list', 'dict', 'OrderedDict']:
                fd = open(self.outfile + '.py', 'w')
                fd.write(str(self.data))
            elif self.result == 'JSON':
                fd = open(self.outfile + '.json', 'w')
                data = json.loads(self.data, object_pairs_hook=OrderedDict)
                json.dump(data, fd)
            elif self.result == 'XML':
                fd = open(self.outfile + '.xml', 'w')
                root = ET.fromstring(self.data)
                tree = ET.ElementTree(root)
                tree.write(fd, encoding='UTF-8')
            fd.close()
        else:
            return self.data


def fedit(data, title="", comment="", icon=None, parent=None, apply=None,
          ok=True, cancel=True, result='list', outfile=None, type='form',
          scrollbar=False, background_color=None, widget_color=None):
    """
    Create form dialog and return result
    (if Cancel button is pressed, return None)

    :param tuple data: datalist, datagroup (see below)
    :param str title: form title
    :param str comment: header comment
    :param QIcon icon: dialog box icon
    :param QWidget parent: parent widget
    :param str ok: customized ok button label
    :param str cancel: customized cancel button label
    :param tuple apply: (label, function) customized button label and callback
    :param function apply: function taking two arguments (result, widgets)
    :param str result: result serialization ('list', 'dict', 'OrderedDict',
                                             'JSON' or 'XML')
    :param str outfile: write result to the file outfile.[py|json|xml]
    :param str type: layout type ('form' or 'questions')
    :param bool scrollbar: vertical scrollbar
    :param str background_color: color of the background
    :param str widget_color: color of the widgets

    :return: Serialized result (data type depends on `result` parameter)
    
    datalist: list/tuple of (field_name, field_value)
    datagroup: list/tuple of (datalist *or* datagroup, title, comment)
    
    Tips:
      * one field for each member of a datalist
      * one tab for each member of a top-level datagroup
      * one page (of a multipage widget, each page can be selected with a 
        combo box) for each member of a datagroup inside a datagroup
       
    Supported types for field_value:
      - int, float, str, unicode, bool
      - colors: in Qt-compatible text form, i.e. in hex format or name (red,...)
                (automatically detected from a string)
      - list/tuple:
          * the first element will be the selected index (or value)
          * the other elements can be couples (key, value) or only values
    """
    # Create a QApplication instance if no instance currently exists
    # (e.g. if the module is used directly from the interpreter)
    test_travis = os.environ.get('TEST_CI_WIDGETS', None)
    if test_travis is not None:
        app = QApplication.instance()
        if app is None:
            app = QApplication([])
        timer = QTimer(app)
        timer.timeout.connect(app.quit)
        timer.start(1000)
    elif QApplication.startingUp():
        _app = QApplication([])
        translator_qt = QTranslator()
        translator_qt.load('qt_' + QLocale.system().name(),
                       QLibraryInfo.location(QLibraryInfo.TranslationsPath))
        _app.installTranslator(translator_qt)

    serial = ['list', 'dict', 'OrderedDict', 'JSON', 'XML']
    if result not in serial:
        print("Warning: '%s' not in %s, default to list" %
              (result, ', '.join(serial)), file=sys.stderr)
        result = 'list'

    layouts = ['form', 'questions']
    if type not in layouts:
        print("Warning: '%s' not in %s, default to form" %
              (type, ', '.join(layouts)), file=sys.stderr)
        type = 'form'

    dialog = FormDialog(data, title, comment, icon, parent, apply, ok, cancel,
                        result, outfile, type, scrollbar, background_color,
                        widget_color)
    if dialog.exec_():
        return dialog.get()


if __name__ == "__main__":

    def create_datalist_example():
        return [('str', 'this is a string'),
                ('str', """this is a 
                MULTILINE
                string"""),
                ('list', [0, '1', '3', '4']),
                ('list2', ['--', ('none', 'None'), ('--', 'Dashed'),
                           ('-.', 'DashDot'), ('-', 'Solid'),
                           ('steps', 'Steps'), (':', 'Dotted')]),
                ('float', 1.2),
                (None, 'Other:'),
                ('int', 12),
                ('font', ('Arial', 10, False, True)),
                ('color', '#123409'),
                ('bool', True),
                ('date', datetime.date(2010, 10, 10)),
                ('datetime', datetime.datetime(2010, 10, 10)),
                ]
        
    def create_datagroup_example():
        datalist = create_datalist_example()
        return ((datalist, "Category 1", "Category 1 comment"),
                (datalist, "Category 2", "Category 2 comment"),
                (datalist, "Category 3", "Category 3 comment"))
    
    #--------- datalist example
    datalist = create_datalist_example()
    def apply_test(data):
        print("data:", data)
    print("result:", fedit(datalist, title="Example",
                           comment="This is just an <b>example</b>.",
                           apply=apply_test))
    
    #--------- datagroup example
    datagroup = create_datagroup_example()
    print("result:", fedit(datagroup, "Global title"))
    
    #--------- datagroup inside a datagroup example
    datalist = create_datalist_example()
    datagroup = create_datagroup_example()
    print("result:", fedit(((datagroup, "Title 1", "Tab 1 comment"),
                            (datalist, "Title 2", "Tab 2 comment"),
                            (datalist, "Title 3", "Tab 3 comment")),
                            "Global title"))



import io
import sys
import threading
import time
import logging
from logging.handlers import RotatingFileHandler
from logging import StreamHandler, Handler
from queue import Queue

from .conf import config
from .gui_proxy import gui

sentinel = object()

logger = logging.getLogger(__name__)

ESC = '\033['

RED_PREFIX = ESC + '38;5;9m'
RED_SUFFIX = ESC + '0m'

LOG_PREFIX = '\033[48;5;7mLOG\033[0m '

DEBUG_PREFIX = ESC + '38;5;6m'
DEBUG_SUFFIX = ESC + '0m'
INFO_PREFIX = ESC + '38;5;12m'
INFO_SUFFIX = ESC + '0m'
WARNING_PREFIX = ESC + '38;5;13m'
WARNING_SUFFIX = ESC + '0m'
ERROR_PREFIX = ESC + '38;5;9m'
ERROR_SUFFIX = ESC + '0m'
CRITICAL_PREFIX = ESC + '38;5;5m'
CRITICAL_SUFFIX = ESC + '0m'


if config.get('qapp', False):        
    filehandler = RotatingFileHandler(f'stderr.log', maxBytes=1024*1024, encoding='UTF-8',backupCount=5)
    filehandler.setLevel(config.get('logging_level_logfile', 'DEBUG'))    
    logging.root.addHandler(filehandler)
    
    
class StreamRouter(object):
    '''
    Pass the calls to a threading dependent stream
    '''
    
    def __init__(self):
        self.streams = dict()
        self.backup_stream = sys.__stdout__
    
    @property
    def stream(self):
        ident = threading.get_ident()
        
        if ident in self.streams.keys():       
            return self.streams[ident]
            
        else:
            return self.backup_stream
    
    def route_stream(self, stream, ident=None):  
        if ident is None:
            ident = threading.get_ident()
        self.streams[ident] = stream
        
    def unregister(self, ident=None):  
        if ident is None:
            ident = threading.get_ident()
        self.streams.pop(ident)
        
    def copy_to_thread(self, to_tid, from_tid=None):
        from_tid = from_tid or threading.get_ident()
        self.streams[to_tid] = self.streams[from_tid]
        
    def __getattr__(self, attr):
        return getattr(self.stream, attr)
        
    def __dir__(self):
        return dir(self.stream)
            
            
class StdOutRouter(StreamRouter):
    def __init__(self):
        super().__init__()
        self.backup_stream = sys.__stdout__
    
    
class StdErrRouter(StreamRouter):
    def __init__(self):
        super().__init__()
        self.backup_stream = sys.__stderr__
        
        
class GhStreamHandler(Handler):
    def __init__(self, stream):
        super().__init__()
        self.stream = stream
        self.set_name('ghstream')
        
    def setStream(self, stream):
        self.stream.flush()
        self.stream = stream

    def emit(self, record):
        text = self.format(record)               

        if record.levelno <= logging.DEBUG:
            self.stream.write(f'{LOG_PREFIX}{DEBUG_PREFIX}{text}{DEBUG_SUFFIX}\n')
        elif record.levelno <= logging.INFO:
            self.stream.write(f'{LOG_PREFIX}{INFO_PREFIX}{text}{INFO_SUFFIX}\n')            
        elif record.levelno <= logging.WARNING:
            self.stream.write(f'{LOG_PREFIX}{WARNING_PREFIX}{text}{WARNING_SUFFIX}\n')
        elif record.levelno <= logging.ERROR:
            self.stream.write(f'{LOG_PREFIX}{ERROR_PREFIX}{text}{ERROR_SUFFIX}\n')
        elif record.levelno <= logging.CRITICAL:
            self.stream.write(f'{LOG_PREFIX}{CRITICAL_PREFIX}{text}{CRITICAL_SUFFIX}\n')            
        else:        
            self.stream.write(f'LOG {text}')

        try:
            if record.levelno >= logging.WARNING:
                gui.console.show_me()
        except:
            pass


class PopupHandler(Handler):
    def emit(self, record):
        #Do not popup outside the main gui thread
        #Should it not be better to test console thread ?
        #Otherwise, errors in Qt Threads, will not popup
        if gui.qapp is None: return
        
        text = self.format(record)

        if record.levelno == logging.ERROR:        
            gui.dialog.msgbox(text, 'Error', 'error')            
        elif record.levelno == logging.CRITICAL:
            gui.dialog.msgbox(text, 'Critical', 'error')           

def enable_ghstream_handler():    
    streamhandler = GhStreamHandler(sys.stdout)
    streamhandler.setLevel(config.get('logging_level_console', 'WARNING'))
    logging.root.addHandler(streamhandler)       

class FlushReducer(object):
    def __init__(self, flusher):
        self.q = Queue()
        self.flusher = flusher
        self.thread = threading.Thread(target=self.reduce, name='FlushReducer', daemon=True)
        self.thread.start()
        
    def __call__(self):
        self.q.put(1)
        
    def reduce(self):
        while True:
            t = self.q.get()
            if t is sentinel:
                return
            time.sleep(0.01)
            self.q.queue.clear()
            self.flusher()
            
    def close(self):
        #wait on an empty queue
        while not self.q.empty():
            time.sleep(0.01)
        self.q.put(sentinel)
        
class FlushPipeStream(io.TextIOBase):
    def __init__(self, streamqueue, flusher):
        self.streamqueue = streamqueue
        self.flusher = FlushReducer(flusher)        
        
    def write(self, text):
        self._write_mode(text, config['stdoutmode'])

    def ansi(self, text):
        self._write_mode(text, 'ansi')
        
    def _write_mode(self, text, mode, prefix='', suffix=''):            
        if mode == 'ansi':
            text = f'{prefix}{text}{suffix}'
            
        self.streamqueue.put((mode, text))
        self.flush()               
        
    def flush(self): 
        self.flusher()      

class ErrLogStream(io.TextIOBase):
    def __init__(self):
        self.line_cache = ''      
        
    def write(self, text):        
        self.line_cache += text
        
        if text.endswith('\n'):        
            logger.error(self.line_cache.rstrip('\n'))
            self.line_cache = ''        
        
class ProcessStdInput(io.TextIOBase):   
    stdin_queues = dict()   
    
    def __init__(self, stdin_queue=None):
        self.stdin_queue = stdin_queue
        
    def close(self):
        pass
        
    def read(self):
        ident = threading.get_ident()
        
        if ident in ProcessStdInput.stdin_queues.keys():
            text = ProcessStdInput.stdin_queues[ident].get()                        
            return text
        else:
            return sys.__stdin__.read()


import logging

logger = logging.getLogger(__name__)

class LayoutStruct(object):
    def __init__(self):
        self.root = dict()
        self.tabblock = True  #Don't split inside tabs, go back one level

    def insert_panel(self, panel, relation='right', to_panel=None, size=100):
        category, panid = panel
        new_node = {'type': 'panel', 'category': category, 'id': panid}
        self.insert_panel_tree(self.root, new_node, relation, to_panel, size)

    def insert_branch(self, branch, relation='right', to_panel=None, size=100):
        self.insert_panel_tree(self.root, branch, relation, to_panel, size)

    def is_empty(self):
        def node_is_empty(node):
            if len(node.keys()) == 0:
                return True

            elif node['type'] == 'panel':
                return False

            for node in node.get('items', []):
                if not node_is_empty(node):
                    return False

            return True

        return node_is_empty(self.root)

    def pop_node(self, nodetype='layout', category='tab', node_id=None):

        def check_node(subnode):
            if subnode['type'] == nodetype and \
                subnode['category'] == category and \
                subnode.get('id', None) == node_id:
                return True
            else:
                return False

        if check_node(self.root):
            poped_node = self.root
            self.root = dict()

            l = LayoutStruct()
            l.root = poped_node
            return l

        if not 'items' in self.root.keys():
            return None

        def pop_node_brach(node):
            #print(f'node: {node}')
            found_index = None
            poped_node = None
            for index, subnode in enumerate(node['items']):
                if check_node(subnode):
                    found_index = index
                    break

                elif poped_node is None and subnode['type'] != 'panel':
                    poped_node = pop_node_brach(subnode)

            if poped_node is None and not found_index is None:
                poped_node = node['items'].pop(found_index)
                if 'sizes' in node.keys():
                    pinned_size = len(node['sizes'])
                    if found_index < pinned_size:
                        #In the pinned area
                        node['sizes'].pop(found_index)
                    else:
                        #In the scroll area
                        node['scroll'].pop(found_index-pinned_size)

            #print(f'poped_node: {poped_node}')
            return poped_node

        poped_root = pop_node_brach(self.root)
        l = LayoutStruct()
        l.root = poped_root
        return l

    def insert_panel_tree(self, node, new_node, relation, to_panel=None, size=100):
        """
        Relation is either: 'left', 'right', 'top', 'bottom' or 'tab'
        """

        if not to_panel is None and not node['type'] == 'panel':
            top = False
            to_category, to_panid = to_panel
            item_index = None
            found_deeper = False
            for index, item in enumerate(node['items']):
                if item['type'] == 'panel':
                    if item['category'] == to_category and item['id'] == to_panid:
                        item_index = index
                else:
                    found_deeper = self.insert_panel_tree(item, new_node, relation, to_panel, size)
                    if found_deeper:
                        item_index = index

        else:
            top = True
            if relation in ['left', 'top']:
                item_index = 0

            elif relation in ['right', 'bottom']:
                if node['type'] == 'panel':
                    #probably as top level single panel
                    item_index = -1
                else:
                    item_index = len(node['items']) - 1

            elif relation in ['tab']:
                item_index = -1 #Will not be used

        if not item_index is None:
            if relation in ['tab']:
                if not node['category'] == 'tab':
                    if top:
                        self.root = {'type': 'layout', 'category': 'tab', 'items': [node]}
                        self.insert_panel_tree(self.root, new_node, relation, to_panel, size)
                    else:
                        subnode = node['items'][item_index]
                        node['items'][item_index] = {'type': 'layout', 'category': 'tab', 'items': [subnode]}
                        self.insert_panel_tree(node['items'][item_index], new_node, relation, to_panel, size)
                else:
                    node['items'].append(new_node)

            else:
                if relation in ['left', 'right']:
                    boxtype = 'hbox'

                elif relation in ['top', 'bottom']:
                    boxtype = 'vbox'

                if not node['category'] == boxtype:
                    if top:
                        #insert a box at top level
                        self.root = {'type': 'layout', 'category': boxtype, 'items': [node], 'sizes': [size]}
                        self.insert_panel_tree(self.root, new_node, relation, to_panel, size)
                    elif self.tabblock and node['category'] == 'tab':
                        #Go back one level
                        return True
                    else:
                        #insert a box at this level
                        subnode = node['items'][item_index]
                        node['items'][item_index] = {'type': 'layout', 'category': boxtype, 'items': [subnode], 'sizes': [size]}
                        self.insert_panel_tree(node['items'][item_index], new_node, relation, to_panel, size)

                elif node['category'] == boxtype:
                    #Add the item to the current hbox
                    first_scroll_index = len(node['sizes'])
                    if relation in ['left', 'top']:
                        if item_index < first_scroll_index:
                            node['items'].insert(item_index, new_node)
                            node['sizes'].insert(item_index, size)
                        else:
                            node['items'].insert(item_index, new_node)
                            node['scroll'].insert(item_index-first_scroll_index, size)

                    elif relation in ['right', 'bottom']:
                        if item_index < first_scroll_index:
                            node['items'].insert(item_index+1, new_node)
                            node['sizes'].insert(item_index+1, size)
                        else:
                            node['items'].insert(item_index+1, new_node)
                            node['scroll'].insert(item_index-first_scroll_index+1, size)


    def compact(self):
        if not 'items' in self.root.keys():
            return

        self.compact_branch(self.root)

        if len(self.root['items']) == 0:
            self.root = dict()
            self.root['type'] = 'layout'
            self.root['category'] = 'hbox'
            self.root['items'] = []
            self.root['sizes'] = []

        elif len(self.root['items']) == 1:
            self.root = self.root['items'][0]

    def compact_branch(self, node):
        zero_items = []

        for index, subnode in enumerate(node['items']):
            if subnode['type'] == 'panel':
                continue

            self.compact_branch(subnode)

            if len(subnode['items']) == 0:
                zero_items.append(subnode)

            elif len(subnode['items']) == 1:
                node['items'][index] = subnode['items'][0]

        for item in zero_items:
            node['items'].remove(item)

    def distribute(self):
        def distribute_branch(node):
            if node['type'] == 'panel':
                return

            if 'sizes' in node.keys():
                mean = sum(node['sizes']) / len(node['sizes'])
                node['sizes'] = len(node['sizes']) * [mean]

            for subnode in node['items']:
                distribute_branch(subnode)

        distribute_branch(self.root)

    def show(self):
        print(self.describe())

    def describe(self):
        if self.root.get('type', None) == None:
            return

        lines = LayoutStruct.show_branch(self.root)
        return '\n'.join(lines)

    @staticmethod
    def show_branch(node, prefix=''):
        lines = []

        if node['type'] == 'panel':
            lines.append(f"{prefix}{node['category']}#{node['id']}")

        elif node['category'] in ['tab', 'tag']:
            tabid = node.get('id', None)
            active = node.get('active', '')
            lines.append(f"{prefix}{node['type']} {node['category']} {tabid} {active}")
            for subnode in node['items']:
                lines.extend(LayoutStruct.show_branch(subnode, prefix + '  '))

        else:
            tabid = node.get('id', None)
            lines.append(f"{prefix}{node['type']} {node['category']} {tabid} {node['sizes']}")
            for subnode in node['items']:
                lines.extend(LayoutStruct.show_branch(subnode, prefix + '  '))

        return lines


"""The is the doctring of live.
Example of nested live script

-- scripts/info.py ---------------------

 def get_info():
     return f'Info from {__file__}'

----------------------------------------

-- scripts/map1/hello.py ---------------

 from gdesk.live import using
 info = using.info

 def hello_world():
     print(f'Hello world')
     print(info.get_info())

----------------------------------------

-- Use of using at top level -----------

 from gdesk.live import manager, use
 manager.append_path('scripts')
 hello = use.map1.hello
 hello.hello_world() 
"""

import sys
import os

from .manage import LiveScriptManager, LiveScriptScan

workspace = dict()

manager = LiveScriptManager(workspace)        

PATH_SEPERATOR = ';' if sys.platform == 'win32' else ':'

if 'LIVEPATH' in os.environ.keys():
    for path in os.environ['LIVEPATH'].split(PATH_SEPERATOR):
        manager.append_path(path)

using = LiveScriptScan(manager, top='__main__')  #Become top if current workspace name is __main__
using_sub = LiveScriptScan(manager, top=False)
use = using_top = LiveScriptScan(manager, top=True)


import pickle

import pylab

from ...core.gui_proxy import GuiProxyBase, StaticGuiCall, gui

class FigureBox(object):
    def __init__(self, figure):
        self.figure = figure
        
    def __getstate__(self):
        state = dict()
        mgr = self.figure.canvas.manager
        self.figure.canvas.manager = None
        state['figure'] = pickle.dumps(self.figure)
        self.figure.canvas.manager = mgr
        return state
        
    def __setstate__(self, state):
        self.figure = pickle.loads(state['figure'])  
        
class PlotProxy(object):
    def __init__(self, gui):
        self.gui = gui        
        
    def __dir__(self):        
        return dir(pylab)
        
    def __getattr__(self, attr):
        func = getattr(pylab, attr)
        return lambda *args, **kwargs: self.gui.gui_call(func, *args, **kwargs)          
       
class PlotGuiProxy(GuiProxyBase):    
    category = 'plot'
    
    def __init__(self):
        pass
        
    def attach(self, gui):
        gui.plot = self    
        gui.prepareplot = self.prepareplot
        self.plx = PlotProxy(gui) 
          
    def show(self, figure=None, hold=False):    
        import matplotlib.pyplot as plt           
        
        from matplotlib.figure import Figure        

        if not gui._qapp is None:
            if figure is None:
                plt.show()
            else:
                plt.figure(figure.number)
                plt.show()
            return
            
        if figure is None:
            figure = plt.gcf()            
       
        assert isinstance(figure, Figure)
        return self.show_figure_box(FigureBox(figure), hold) 
        
    def prepareplot(self):
        return self.plx    

    @StaticGuiCall
    def select(panid=-1, args=()):
        """
        If panid < 0, -1: select the active panel, -2: selected before that, ...
        panid == None: new panel
        panid >= 0: select the panel if exists, otherwise a new with that number
        """    
        panel = gui.qapp.panels.select_or_new('plot', panid, 'basic', args=args)
        return panel.panid
        
    @StaticGuiCall
    def restore_gcf_set_active():
        from . import plotpanel
        plotpanel.restore_gcf_set_active()

    @StaticGuiCall
    def new(figure_id=None):
        """
        Create a new figure.
        
        Links to pylab.figure
        """ 
        import matplotlib.pyplot as plt
        fig = plt.figure(figure_id)
        return fig, fig.number 

    def xy(self, *args, **kwargs):
        if not kwargs.get('hold', False):
            self.plx.figure()
        if 'hold' in kwargs.keys():
            kwargs.pop('hold')
        self.plx.plot(*args, **kwargs)
        self.plx.show()

    @StaticGuiCall  
    def show_figure_box(figurebox, hold=False):
        """
        Display the figure in the figurebox.
        The figure has to be added to the current pyplot backend.            
        
        :param bool hold: Replace the figure in the current selected panel by the new figure
        """
        #unpickling a figure will call show when interactive is on
        #But this will be on the backend of the pickling process 
        #All callbacks related to gui calls where removed by the pickling
        #
        
        from ...matplotbe import new_figure_manager_given_figure
        
        import matplotlib.pyplot as plt
        import matplotlib._pylab_helpers as pylab_helpers
                
        fig = figurebox.figure 

        def make_active(event):
            pylab_helpers.Gcf.set_active(mgr)                     
        
        if False:
            #Creating a new figure manager
            allnums = plt.get_fignums()
            num = max(allnums) + 1 if allnums else 1      
            
            mgr = new_figure_manager_given_figure(num, fig)                                    
            mgr._cidgcf = mgr.canvas.mpl_connect('button_press_event',
                                                     make_active)
                                      
            pylab_helpers.Gcf.set_active(mgr)
            fig = mgr.canvas.figure
            fig.number = num
            
            mgr.show()
        
        else:
            #Use the current figure mamaner and plot panel
            # Try to replace the figure on the current plot panel with this new figure
            from ...matplotbe import FigureCanvasGh2
            panids = gui.get_panel_ids('plot')                       
            
            if not hold or len(panids) == 0:
                ignore, num = PlotGuiProxy.new()
            else:
                num = panids[-1]
            
            plotpanel = gui.qapp.panels['plot'][num]
            fig.number = num            
            plotpanel.showNewFigure(fig)                        
            
            #Hack to call the correct set_active
            #Note that Gcf.set_active() is mapped to panel.select() in plotpanel.py
            mgr = fig.canvas.manager
            mgr._cidgcf = mgr.canvas.mpl_connect('button_press_event', make_active) 
            
            plotpanel.show()
        
        # I don't get the interactive update working
        # After a function lik grid(), xlabel(), ..
        # the stale property is set to  True
        # This should trigger the call of stale_callback on the figure (=Artist)
        # or stall_callback on the axis object
        
        #Configure stale call backs
        #The hierarchy of the figure can be type of figure dependend        
        try:
            from matplotlib.pyplot import _auto_draw_if_interactive
            from matplotlib.figure import _stale_figure_callback
            from matplotlib.artist import _stale_axes_callback
        
            fig.stale_callback = _auto_draw_if_interactive
        
            for ax in fig.axes:
                ax.stale_callback = _stale_figure_callback
                ax.xaxis.stale_callback = _stale_axes_callback
                ax.yaxis.stale_callback = _stale_axes_callback
        except:
            logger.error('Could not configure all interactive callbacks')
            
        return num    
    
    @StaticGuiCall
    def save(file):
        """
        Save the selected figure to a file
        """
        import matplotlib.pyplot as plt
        plt.savefig(file)


"""
Gamma Desk interface to the DOS console
"""
import sys
import logging
import argparse

from . import __release__, refer_shell_instance
from . import configure, config, DOC_HTML, PROGNAME
from .core import conf

logger = logging.getLogger(__name__)

boot_handler = logging.StreamHandler(sys.stdout)
boot_handler.set_name('boot')
logging.root.addHandler(boot_handler)

MODNAME = '.'.join(globals()['__name__'].split('.')[:-1])
PATH_SEPERATOR = ';' if sys.platform == 'win32' else ':'

HEADER = f"{PROGNAME} {__release__}"
HEADER += '\n' + len(HEADER) * '=' + '\n'

HEADER += DOC_HTML + '\n'

EPILOG = f"""\
Examples
--------

{MODNAME} -i init_file.py
{MODNAME} -c config_file.???
"""

def argparser():
    """
    Make the ArgumentParser instance
    """
    parser = argparse.ArgumentParser(description=HEADER, prog=f'python -m {MODNAME}',
        formatter_class=argparse.RawDescriptionHelpFormatter, epilog=EPILOG)

    parser.add_argument("-c", "--config_file", help="Use this configuration file")
    parser.add_argument("-i", "--init_file", help="Run this init file in console 1")
    parser.add_argument("-d", "--debug", action='store_true', help="Set logging level to debug")
    parser.add_argument("pictures", nargs='*', help="Image file to load")

    return parser


def argexec(argv=None, **config_kwargs):
    """
    Configure and start the eventloop
    """
    parser = argparser()
    args = parser.parse_args(argv)

    if args.debug:
        config_kwargs['logging_level'] = 'DEBUG'
        logging.root.setLevel(config_kwargs['logging_level'])

    config_kwargs['qapp'] = True

    if args.config_file:
        config_kwargs['config_file'] = args.config_file

    if args.init_file:
        config_kwargs['init_file'] = args.init_file

    configure(**config_kwargs)

    # Configure has to be done before import other modules
    from .core.shellmod import Shell
    shell = Shell()
    refer_shell_instance(shell)

    from .gcore.guiapp import eventloop
    eventloop(shell, init_file=config['init_file'], pictures=args.pictures)

    return shell


def run_as_child(console_args, config_kwargs, config_objects):
    """
    Top level function to start as child process
    """
    #Note that auto unpickling of received arguments can have caused a configarion to be execed
    #The configuration was triggered by the Process code on decode this function pointer
    conf.config_objects.update(config_objects)

    #Allow reconfiguring
    conf.config.clear()
    conf.configured = False

    print(config_kwargs)

    argexec(console_args, **config_kwargs)


def is_imported_by_child_process():
    """
    Detect whenever this module is imported by multiprocessing.spawn
    """
    frame = sys._getframe()

    while not frame is None:
        module_name = frame.f_globals['__name__']
        if module_name == 'multiprocessing.spawn':
            return True
        frame = frame.f_back

    return False


if is_imported_by_child_process():
    configure(qapp=True)



from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Qt

class Thumbs(QtWidgets.QListWidget):

    def __init__(self, parent):
        super().__init__(parent=parent)
        #self.setViewMode(QtWidgets.QListWidget.IconMode)
        self.setViewMode(QtWidgets.QListWidget.ListMode)
        self.setIconSize(QtCore.QSize(256, 256))
        self.setResizeMode(QtWidgets.QListWidget.Adjust)

    def addQImage(self, qimg, title):
        pixmap = QtGui.QPixmap(qimg)
        self.addItem(QtWidgets.QListWidgetItem(QtGui.QIcon(pixmap), title))


from ...core.gui_proxy import GuiProxyBase, StaticGuiCall, gui
        
class HtmlGuiProxy(GuiProxyBase):    
    category = 'html'
    
    def __init__(self):
        pass
        
    def attach(self, gui):
        gui.html = self
    
    @StaticGuiCall
    def show(html_content):
        #Note that the size of content is limited
        panel = gui.qapp.panels.selected('html')
        if panel is None:
            panel = gui.qapp.panels.select_or_new('html')        
        panel.webview.setHtml(html_content)
        
    @StaticGuiCall
    def load_url(url):
        panel = gui.qapp.panels.selected('html')
        if panel is None:
            panel = gui.qapp.panels.select_or_new('html')        
        panel.webview.load(url)    



"""The Garbage Collector timer in the Qt Event Loop."""
import gc
import sys

from qtpy.QtCore import QObject, QTimer

class QGarbageCollector(QObject):
    """
    Disable automatic garbage collection and instead collect manually.

    Timeout every INTERVAL milliseconds.

    This is done to ensure that garbage collection only happens in the GUI
    thread, as otherwise Qt can crash.
    """

    INTERVAL = 5000

    def __init__(self, parent, debug=False):
        """QGarbageCollector."""
        QObject.__init__(self, parent)
        self.debug = debug

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.check)

        self.threshold = gc.get_threshold()

    def enable(self):
        """Enable the timer."""
        gc.disable()
        self.timer.start(self.INTERVAL)

    def disable(self):
        """Disable the timer."""
        self.timer.stop()
        gc.enable()

    def check(self):
        """Do the garbage collection."""
        cnt_0, cnt_1, cnt_2 = gc.get_count()
        if self.debug:
            sys.__stdout__.write('gc_check called: %d, %d, %d\n' % (cnt_0, cnt_1, cnt_2))
            sys.__stdout__.flush()

        if cnt_0 > self.threshold[0]:
            num = gc.collect(0)
            if self.debug:
                sys.__stdout__.write('collecting gen 0, found: %d unreachable\n' % num)
                sys.__stdout__.flush()
            if cnt_1 > self.threshold[1]:
                num = gc.collect(1)
                if self.debug:
                    sys.__stdout__.write('collecting gen 1, found: %d unreachable\n' % num)
                    sys.__stdout__.flush()
                if cnt_2 > self.threshold[2]:
                    num = gc.collect(2)
                    if self.debug:
                        sys.__stdout__.write('collecting gen 2, found: %d unreachable\n' % num)
                        sys.__stdout__.flush()



import numpy
import time 
import math

HFG = 8
USEFORDER = True
XKERN = 8
YKERN = 8

def make_thumbnail(array, max_long_side=240, hfg=HFG, bayer=False):

    if hasattr(array, 'tondarray'):
        array = image.tondarray('uint16')  # reference to uint16 buffer of Bindex 

    height, width = array.shape
    long_side = max(height, width)
    ratio = math.ceil(long_side / max_long_side)

    thumb = make_blueprint(array, ratio, ratio, hfg, bayer)
    return thumb
    
def get_blue_print(array, xkern=XKERN, ykern=YKERN, hfg=HFG, bayer=False):
    return make_blueprint(array, xkern, ykern, hfg, bayer)

def make_blueprint(array, xkern=XKERN, ykern=YKERN, hfg=HFG, bayer=False):
    """
    Make a so called blueprint thumbnail.
    The 16-bitgrey image is converted to a downscaled color version.
    The downscale kernel is tyical 8x8. So the statiscs of 64 pixels is used
    to derive 3 8-bit numbers. So the RGB values are based on the average,
    the minimum and the maximum.
    Green for the average of the kernel.
    Blue for the difference of the minumum to average.
    Red for the difference of maximum to average.
    The min-max differences are typical gained up.
    The final result is an image of 1/100 of the original data.
    It blows-up high frequency artifacts, making them much more visible.
    
    :param array: A 2d Image
    :type array: np.ndarray or PyBix
    :param int xkern: width of the kernel
    :param int ykern: height of the kernel
    :param float hfg: high frequency gain        
    :param bool bayer: Does it contains a bayer pattern (color sensor?)
    """

    def get_blue_print_of_roi(array, xkern, ykern, hfg):
        #the shape of array is supposed to be a multiple of xkern and ykern
        
        ydim, xdim = array.shape
        xdimsc = xdim // xkern
        ydimsc = ydim // ykern
        
        #note that a reshape doesn't realy do anything on the memory buffer
        #it only redefines the strides
        tmp2 = array.reshape(ydim * xdimsc, xkern)
        
        #calculates on every xkern succesive pixels in a row
        #total size of the array will be 8 times less
        min3 = tmp2.min(1)
        mean3 = tmp2.mean(1)
        max3 = tmp2.max(1)    
        
        #reshufle the memory
        #the copy is needed to really remap the memory
        if USEFORDER == False:
            #swap row and column in memory
            min4 = min3.reshape(ydim, xdimsc).T.copy()
            mean4 = mean3.reshape(ydim, xdimsc).T.copy()
            max4 = max3.reshape(ydim, xdimsc).T.copy()   
        
            min5 = min4.reshape(xdimsc * ydimsc, ykern)
            mean5 = mean4.reshape(xdimsc * ydimsc, ykern)
            max5 = max4.reshape(xdimsc * ydimsc, ykern)
            
            stataxis = 1
                    
        else:
            #the use or order='F' actual makes a copy
            #so we don't gain anything with this Fortran order
            min5 = min3.reshape(ydim, xdimsc).reshape(ykern, xdimsc * ydimsc, order='F')
            mean5 = mean3.reshape(ydim, xdimsc).reshape(ykern, xdimsc * ydimsc, order='F')
            max5 = max3.reshape(ydim, xdimsc).reshape(ykern, xdimsc * ydimsc, order='F')
            
            stataxis = 0
            
        #calculates on every 8 succesive pixels in a col
        #total size of the array will be now 64 times less then the orginal
        min6 = min5.min(stataxis)
        mean6 = mean5.mean(stataxis)
        max6 = max5.max(stataxis)        
            
        blueprint = numpy.ndarray((ydimsc, xdimsc, 3),'uint8')    
        
        if array.dtype == 'uint8':
            scale = 1
        elif array.dtype == 'uint16':
            scale = 256
            
        min7 = min6.reshape(xdimsc, ydimsc).T[:,:] // scale
        mean7 = mean6.reshape(xdimsc, ydimsc).T[:,:] // scale
        max7 = max6.reshape(xdimsc, ydimsc).T[:,:] // scale     
        
        blueprint[:,:,2] = (255 - (mean7 - min7) * hfg).clip(0,255)
        blueprint[:,:,1] = mean7
        blueprint[:,:,0] = ((max7 - mean7) * hfg).clip(0,255)
        
        #print("it took %8.2f seconds " % (time.time() - timestamp))
        return blueprint
        
    def blue_reduce_mono(array, kernel_width, kernel_height, hfgain):
        """
        array is supposed to be debayered
        """
    
        ydim, xdim = array.shape
    
        #round down and up to multiple of kernel widths and height
        xroundlow, xroundhigh = xdim // kernel_width * kernel_width, -(-xdim // kernel_width * kernel_width)
        yroundlow, yroundhigh = ydim // kernel_height * kernel_height, -(-ydim // kernel_height * kernel_height)
    
        if xroundlow == xroundhigh and yroundlow == yroundhigh:
            blue_print =  get_blue_print_of_roi(array, kernel_width, kernel_height, hfgain)
            
        else:
            xdimsc = xroundlow // xkern
            ydimsc = yroundlow // ykern
            if xroundlow == xroundhigh:          
                blue_print = numpy.ndarray((ydimsc+1, xdimsc, 3),'uint8')
                blue_print[:ydimsc,:] =  get_blue_print_of_roi(array[:yroundlow, :], xkern, ykern, hfgain)
                blue_print[ydimsc:,:] =  get_blue_print_of_roi(array[yroundlow:, :], xkern, ydim-yroundlow, hfgain)
            elif yroundlow == yroundhigh:
                blue_print = numpy.ndarray((ydimsc, xdimsc+1, 3),'uint8')
                blue_print[:, :xdimsc] =  get_blue_print_of_roi(array[:, :xroundlow], xkern, ykern, hfgain)
                blue_print[:, xdimsc:] =  get_blue_print_of_roi(array[:, xroundlow:], xdim-xroundlow, ykern, hfgain)        
            else:
                blue_print = numpy.ndarray((ydimsc+1, xdimsc+1, 3),'uint8')
                blue_print[:ydimsc, :xdimsc] =  get_blue_print_of_roi(array[:yroundlow, :xroundlow], xkern, ykern, hfgain)
                blue_print[:ydimsc, xdimsc:] =  get_blue_print_of_roi(array[:yroundlow, xroundlow:], xdim-xroundlow, ykern, hfgain)
                blue_print[ydimsc:, :xdimsc] =  get_blue_print_of_roi(array[yroundlow:, :xroundlow], xkern, ydim-yroundlow, hfgain)
                blue_print[ydimsc:, xdimsc:] =  get_blue_print_of_roi(array[yroundlow:, xroundlow:], xdim-xroundlow, ydim-yroundlow, hfgain)            
                
        return blue_print

    if hasattr(array, 'tondarray'):
        array = image.tondarray('uint16')  # reference to uint16 buffer of Bindex            
        
    if not len(array.shape) == 2:
        raise AttributeError('Only images with 1 channel (mono) is supported')
            
    if not bayer:
        blue_print = blue_reduce_mono(array, xkern, ykern, hfg)
        
    else:
        blue_prints = []
        for i,j in ((0,0),(1,0),(0,1),(1,1)):
            blue_prints.append(blue_reduce_mono(array[i::2,j::2], xkern, ykern, hfg))
            
        blue_print_width = blue_prints[0].shape[1] + blue_prints[2].shape[1]
        blue_print_height = blue_prints[0].shape[0] + blue_prints[1].shape[0]
        blue_print = numpy.zeros((blue_print_height, blue_print_width, 3), 'uint8')
        
        substarts = [(0,0),(1,0),(0,1),(1,1)]
        for i, sub_blue_print in enumerate(blue_prints):
            j, k = substarts[i]
            blue_print[j::2,k::2] = sub_blue_print
            
    return blue_print


import ctypes
from ctypes import wintypes
import time

user32 = ctypes.WinDLL('user32', use_last_error=True)

INPUT_MOUSE    = 0
INPUT_KEYBOARD = 1
INPUT_HARDWARE = 2

KEYEVENTF_EXTENDEDKEY = 0x0001
KEYEVENTF_KEYUP       = 0x0002
KEYEVENTF_UNICODE     = 0x0004
KEYEVENTF_SCANCODE    = 0x0008

MAPVK_VK_TO_VSC = 0

# msdn.microsoft.com/en-us/library/dd375731
VK_TAB  = 0x09
VK_MENU = 0x12
VK_F15 = 0x7E

# C struct definitions

wintypes.ULONG_PTR = wintypes.WPARAM

class MOUSEINPUT(ctypes.Structure):
    _fields_ = (("dx",          wintypes.LONG),
                ("dy",          wintypes.LONG),
                ("mouseData",   wintypes.DWORD),
                ("dwFlags",     wintypes.DWORD),
                ("time",        wintypes.DWORD),
                ("dwExtraInfo", wintypes.ULONG_PTR))

class KEYBDINPUT(ctypes.Structure):
    _fields_ = (("wVk",         wintypes.WORD),
                ("wScan",       wintypes.WORD),
                ("dwFlags",     wintypes.DWORD),
                ("time",        wintypes.DWORD),
                ("dwExtraInfo", wintypes.ULONG_PTR))

    def __init__(self, *args, **kwds):
        super(KEYBDINPUT, self).__init__(*args, **kwds)
        # some programs use the scan code even if KEYEVENTF_SCANCODE
        # isn't set in dwFflags, so attempt to map the correct code.
        if not self.dwFlags & KEYEVENTF_UNICODE:
            self.wScan = user32.MapVirtualKeyExW(self.wVk,
                                                 MAPVK_VK_TO_VSC, 0)

class HARDWAREINPUT(ctypes.Structure):
    _fields_ = (("uMsg",    wintypes.DWORD),
                ("wParamL", wintypes.WORD),
                ("wParamH", wintypes.WORD))

class INPUT(ctypes.Structure):
    class _INPUT(ctypes.Union):
        _fields_ = (("ki", KEYBDINPUT),
                    ("mi", MOUSEINPUT),
                    ("hi", HARDWAREINPUT))
    _anonymous_ = ("_input",)
    _fields_ = (("type",   wintypes.DWORD),
                ("_input", _INPUT))

LPINPUT = ctypes.POINTER(INPUT)

def _check_count(result, func, args):
    if result == 0:
        raise ctypes.WinError(ctypes.get_last_error())
    return args

user32.SendInput.errcheck = _check_count
user32.SendInput.argtypes = (wintypes.UINT, # nInputs
                             LPINPUT,       # pInputs
                             ctypes.c_int)  # cbSize

# Functions

def PressKey(hexKeyCode):
    x = INPUT(type=INPUT_KEYBOARD,
              ki=KEYBDINPUT(wVk=hexKeyCode))
    try:
        user32.SendInput(1, ctypes.byref(x), ctypes.sizeof(x))
    except:
        # [WinError 5] Access is denied.
        # Can happen when current session is locked 
        # by for example switching to other users
        pass

def ReleaseKey(hexKeyCode):
    x = INPUT(type=INPUT_KEYBOARD,
              ki=KEYBDINPUT(wVk=hexKeyCode,
                            dwFlags=KEYEVENTF_KEYUP))
    try:
        user32.SendInput(1, ctypes.byref(x), ctypes.sizeof(x))
    except:
        # [WinError 5] Access is denied.
        # Can happen when current session is locked 
        # by for example switching to other users        
        pass
    
def red_bulling(interval = 60):
    """
    Press F-15 every interval seconds
    """    
    import time
    print('Red bulling started')
    while True:
        time.sleep(interval)
        PressKey(0x7E)
        ReleaseKey(0x7E)
        
if __name__ == "__main__":
    red_bulling()


"""
Render to qt from agg.
"""
import os
import ctypes
import sys
import threading
import pickle
import logging
import warnings

import matplotlib
from matplotlib.transforms import Bbox
from matplotlib.figure import Figure
from matplotlib import cbook
from matplotlib._pylab_helpers import Gcf
from matplotlib.backend_bases import FigureCanvasBase, FigureManagerBase
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_qt5 import (
    QtCore, QtGui, QtWidgets, _BackendQT5, FigureCanvasQT, FigureManagerQT,
    NavigationToolbar2QT, backend_version)   
  
from matplotlib.backends.qt_compat import QT_API
from matplotlib.backends.backend_template import FigureCanvasTemplate, FigureManagerTemplate

if matplotlib.__version__[:3] in ['3.2', '3.3']:
    from matplotlib.backends.qt_compat import _setDevicePixelRatioF
    setDevicePixelRatio = _setDevicePixelRatioF
    
elif matplotlib.__version__[:3] in ['3.4']:
    from matplotlib.backends.qt_compat import _setDevicePixelRatio
    setDevicePixelRatio = _setDevicePixelRatio
    
else:
    warnings.warn(
        f'Matplotlib version {matplotlib.__version__} not supported.\n'
        f'Version should be 3.2.x, 3.3.x or 3.4.x')
    
from .. import gui
from ..panels.matplot import PlotPanel

logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore", "Starting a Matplotlib GUI outside of the main thread will likely fail.")

def draw_if_interactive():
    """
    For image backends - is not required.
    For GUI backends - this should be overridden if drawing should be done in
    interactive python mode.
    """  
    if matplotlib.is_interactive(): 
        show()

def show(*, block=None):
    """
    For image backends - is not required.
    For GUI backends - show() is usually the last line of a pyplot script and
    tells the backend that it is time to draw.  In interactive mode, this
    should do nothing.
    """ 
    #manager = Gcf.get_active()    
    for manager in Gcf.get_all_fig_managers():
        manager.show()
        
def new_figure_manager(num, *args, FigureClass=Figure, **kwargs):
    """Create a new figure manager instance."""
    # If a main-level app must be created, this (and
    # new_figure_manager_given_figure) is the usual place to do it -- see
    # backend_wx, backend_wxagg and backend_tkagg for examples.  Not all GUIs
    # require explicit instantiation of a main-level app (e.g., backend_gtk3)
    # for pylab.
    thisFig = FigureClass(*args, **kwargs)
    return new_figure_manager_given_figure(num, thisFig)

def new_figure_manager_given_figure(num, figure):
    """Create a new figure manager instance for the given figure."""
    
    #print(f'timer: {time.perf_counter()}')

    if not gui.valid() or gui._qapp is None:                    
        #In case of comming from other Process
        #Don't do a guicall, FigureCanvasGh2 or FigureManagerQT is not pickable!
        #Is called if figure, line, ... is depickled from the interprocess queue
        canvas = FigureCanvasBase(figure)
        manager = FigureManagerGh2Child(canvas, num)
        
    else:
        canvas = gui.gui_call(FigureCanvasGh2, figure)    
        manager = FigureManagerGh2(canvas, num)
        
    return manager        


class FigureCanvasGh2(FigureCanvasAgg, FigureCanvasQT):

    def __init__(self, figure):
        # Must pass 'figure' as kwarg to Qt base class.
        super().__init__(figure=figure)

    def paintEvent(self, event):
        """
        Copy the image from the Agg canvas to the qt.drawable.

        In Qt, all drawing should be done inside of here when a widget is
        shown onscreen.
        """
        logger.debug('calling paintEvent')
        
        if matplotlib.__version__[:3] in ['3.2', '3.3']:
            if self._update_dpi():
                # The dpi update triggered its own paintEvent.
                return
                
        self._draw_idle()  # Only does something if a draw is pending.

        # If the canvas does not have a renderer, then give up and wait for
        # FigureCanvasAgg.draw(self) to be called.
        if not hasattr(self, 'renderer'):
            return

        painter = QtGui.QPainter(self)
        try:
            # See documentation of QRect: bottom() and right() are off
            # by 1, so use left() + width() and top() + height().
            rect = event.rect()
            # scale rect dimensions using the screen dpi ratio to get
            # correct values for the Figure coordinates (rather than
            # QT5's coords)
            width = rect.width() * self._dpi_ratio
            height = rect.height() * self._dpi_ratio
            left, top = self.mouseEventCoords(rect.topLeft())
            # shift the "top" by the height of the image to get the
            # correct corner for our coordinate system
            bottom = top - height
            # same with the right side of the image
            right = left + width
            # create a buffer using the image bounding box
            bbox = Bbox([[left, bottom], [right, top]])
            reg = self.copy_from_bbox(bbox)
            buf = cbook._unmultiplied_rgba8888_to_premultiplied_argb32(
                memoryview(reg))

            # clear the widget canvas
            painter.eraseRect(rect)

            qimage = QtGui.QImage(buf, buf.shape[1], buf.shape[0],
                                  QtGui.QImage.Format_ARGB32_Premultiplied)
            setDevicePixelRatio(qimage, self._dpi_ratio)
            # set origin using original QT coordinates
            origin = QtCore.QPoint(rect.left(), rect.top())
            painter.drawImage(origin, qimage)
            # Adjust the buf reference count to work around a memory
            # leak bug in QImage under PySide on Python 3.
            if QT_API in ('PySide', 'PySide2'):
                ctypes.c_long.from_address(id(buf)).value = 1

            self._draw_rect_callback(painter)
        finally:
            painter.end()
            
    def draw_idle(self):
        logger.debug('calling draw_idle')
        gui.gui_call(FigureCanvasQT.draw_idle, self) 
        
    def destroy(self, *args):
        gui.gui_call(FigureCanvasQT.destroy, self, *args)                                    

    def blit(self, bbox=None):
        # docstring inherited
        # If bbox is None, blit the entire canvas. Otherwise
        # blit only the area defined by the bbox.
        if bbox is None and self.figure:
            bbox = self.figure.bbox

        # repaint uses logical pixels, not physical pixels like the renderer.
        l, b, w, h = [int(pt / self._dpi_ratio) for pt in bbox.bounds]
        t = b + h
        self.repaint(l, self.renderer.height / self._dpi_ratio - t, w, h)

    def print_figure(self, *args, **kwargs):
        super().print_figure(*args, **kwargs)
        self.draw()
        
class FigureManagerGh2(FigureManagerBase):

    """
    Wrap everything up into a window for the pylab interface

    For non interactive backends, the base class does all the work
    """       
    def __init__(self, canvas, num):
        super().__init__(canvas, num)
        self.panel = gui.gui_call(gui._qapp.panels.new_panel, PlotPanel, None, num, None, args=(canvas,))
    
    def show(self):
        """
        For GUI backends, show the figure window and redraw.
        For non-GUI backends, raise an exception to be caught
        by :meth:`~matplotlib.figure.Figure.show`, for an
        optional warning.
        """
        if gui.valid():            
            gui.gui_call(PlotPanel.show, self.panel)
            gui.gui_call(PlotPanel.refresh, self.panel)
        else:        
            raise ValueError(f'gui called from unknown thread {os.getpid()}/{threading.current_thread()}')
        
    def destroy(self, *args):        
        if 'plot' in gui._qapp.panels.keys():        
            gui.gui_call(PlotPanel.close_panel, self.panel)
        else:
            pass
            
class FigureManagerGh2Child(FigureManagerBase):

    """
    Wrap everything up into a window for the pylab interface

    For non interactive backends, the base class does all the work
    """       
    def __init__(self, canvas, num):
        super().__init__(canvas, num)
        self.panel = None
    
    def show(self):
        """
        For GUI backends, show the figure window and redraw.
        For non-GUI backends, raise an exception to be caught
        by :meth:`~matplotlib.figure.Figure.show`, for an
        optional warning.
        """
        pass
              
        

FigureCanvas = FigureCanvasGh2
FigureManager = FigureManagerGh2



import collections
import importlib
import pprint
import logging
import sys

from qtpy.QtWidgets import *
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy import QtCore, QtGui

from .. import gui

from ..panels.base import BasePanel

from .laystruct import LayoutStruct
from .dockwidgets import DockContainer, DockTab, DockTag
from .overlay import DockOverlay, HoverButton

from ..utils.z_order import get_z_values

logger = logging.getLogger(__name__)

class DockManager(object):
    def __init__(self, panels, qapp):
        self.panels = panels
        self.qapp = qapp
        self.containers = dict()
        self.layoutstructs = dict()
        self.overlays = []
        self.newWindow = lambda name, parentName: gui.qapp.newWindow(name, parentName)
        self.deleteWindow = lambda w: gui.qapp.deleteWindow(w)
        self.perspectives = collections.OrderedDict()
        self.bindGroups = dict()
        
        HoverButton.load_icons()
        
    def new_container(self, parent=None, name='main'):
        container = DockContainer(self, parent, name)
        self.containers[name] = container
        return container
        
    def add_button_to_bindgroup(self, category, panid, button):
        if not category in self.bindGroups.keys():
            self.bindGroups[category] = QButtonGroup(self.qapp)
        
        #If old button can still exist
        #PySide2 doesn't seem to overwrite it, so remove it first
        existingButton = self.bindGroups[category].button(panid)
        if not existingButton is None:
            self.bindGroups[category].removeButton(existingButton)
            
        self.bindGroups[category].addButton(button, panid)
        
    def detach_all(self):
        for category, panels in self.panels.items():        
            for panid, panel in panels.items():
                panel.detach()
                
        for name, container in self.containers.items():
            container.compact()
            
    def get_container(self, category, panid):
        for container in self.containers.values():
            if (category, panid) in container.panelIds:
                return container
        return None

    def get_layout_struct_from_containers(self):
        self.layoutstructs.clear()
        for name, container in self.containers.items():
            self.layoutstructs[name] =  container.get_layout_struct()
        return self.layoutstructs
        
    def show_overlays(self, window, insert_node, allWindows=True):
        if allWindows:
            for container in self.containers.values():
                if not container.isVisible() or container.parent() is window: continue
                overlay = DockOverlay(container, tool=False)
                overlay.show_and_insert(window, insert_node)
                self.overlays.append(overlay)
        else:
            container = window.parent().container
            overlay = DockOverlay(container, tool=False)
            overlay.show_and_insert(window, insert_node)
            self.overlays.append(overlay)            
            
        return len(self.overlays)
        
    def drop_in(self, container, hide=True, allWindows=False):
        window = container.parent()
        if hide: window.hide()
        layout = container.get_layout_struct()
        layout.compact()
        self.show_overlays(window, layout.root, allWindows)
            
    def distribute(self):
        for container in self.containers.values():
            container.distribute()           
        
    def hide_overlays(self):
        for overlay in self.overlays:
            overlay.hide()
            overlay.deleteLater() 
            
        self.overlays.clear()               
            
    def new_window_on_panel(self, panel, parentName=None):
        layout = LayoutStruct()
        layout.root = {'type': 'panel', 'category': panel.category, 'id': panel.panid}
        return self.new_window_using_layout(layout, parentName=parentName)
            
    def new_window_using_layout(self, layout, width=640, height=480, parentName=None):
        w = self.newWindow(None, parentName)
        geo = w.geometry()
        geo.setWidth(width)
        geo.setHeight(height)
        w.setGeometry(geo)
        layout.compact()
        w.container.update_layout(layout)
        w.show()
        return w
        
    def get_perspective(self):        
        perspective = dict()
        perspective['panels'] = []
        perspective['windows'] = []
        
        for category in self.panels.keys():            
            for panid, panel in self.panels[category].items():            
                panelinfo = dict()                
                panelinfo['id'] = panid
                panelinfo['category'] = category
                panelinfo['module'] = type(panel).__module__
                panelinfo['qualname'] = type(panel).__qualname__
                #panelinfo['basewindow'] = panel.baseWindowName
                panelinfo['title'] = panel.windowTitle()
                
                bindings = []
                
                for bindcategory, bindpanid in panel.bindings:
                    binding = dict()
                    binding['category'] = bindcategory
                    binding['id'] = bindpanid
                    bindings.append(binding)
                    
                panelinfo['bindings'] = bindings                
                perspective['panels'].append(panelinfo)        
        
        for name, container in self.containers.items():
            #window = container.parent()
            ls = container.get_layout_struct()
            ls.compact()
            visible = container.parent().isVisible()
            perspective['windows'].append({'name': name, 'docks': ls.root, 'visible': visible})
            
        return perspective

    def set_perspective(self, perspective):
        self.detach_all()
        
        panels = perspective['panels']
        
        #Create panels
        for panelinfo in panels:
            category = panelinfo['category']
            panid = panelinfo['id']
            module = panelinfo['module']
            qualname = panelinfo['qualname']            
            #base_window_name = panelinfo['basewindow']
            
            id_exists = self.panels.id_exists(category, panid)
            
            tmp = importlib.import_module(module)
            for attr in qualname.split('.'):
                tmp = getattr(tmp, attr) 

            Cls = tmp           

            if id_exists:
                panel = self.panels[category][panid]
                if isinstance(panel, Cls):
                    continue
                else:
                    # Delete of this existing panel
                    # Or keep reference to it in some old_panels dictionary?
                    # So the user can still decide what to do with it
                    # What to do with the panid, rename it available id not used in this perspective?
                    # What to do with the bindings
                    # Or just remap the conflicting panid to a fresh panid
                    logger.warn(f'Panel {category} {panid} already exists but is of wrong class {type(panel)}')
                    persp_cat_panids = set(tmppanel['id'] for tmppanel in panels if tmppanel['category'] == category)
                    exist_cat_panids = set(self.panels[category].keys())
                    logger.debug(f'Perspective panids of {category}: {persp_cat_panids}')
                    logger.debug(f'Existing panids of {category}: {exist_cat_panids}')
                    continue                    
                     
            panel = self.panels.new_panel(Cls, None, panid, floating=True) 

        #Setup of bindings
        for panelinfo in panels:
            category = panelinfo['category']
            panid = panelinfo['id']            
            panel = self.panels[category][panid]
            for binding in panelinfo['bindings']:
                bindcategory = binding['category']
                bindpanid = binding['id']
                panel.addBindingTo(bindcategory, bindpanid)      
        
        for window in perspective['windows']:            
            ls = LayoutStruct()
            ls.root = window["docks"]
            visible = window.get('visible', True)
            
            #container = self.qapp.windows['main'].container
            winname = window['name']
            if winname in self.qapp.windows.keys():
                window = self.qapp.windows[winname]
            else:
                window = self.newWindow(None, None)
            window.container.update_layout(ls)
            if visible:
                window.show()
                window.raise_()
            else:
                window.hide()            
        
        #Put all the floating pannels together in tabs and one seperate window
        floating_panels = []
        floating_panels = []
        for category in self.panels.keys():
            for panid, panel in self.panels[category].items():
                if panel.parent() is None:
                    node = {'type': 'panel', 'category': category, 'id': panid}
                    floating_panels.append(node)                  
        
        if len(floating_panels) > 0:
            layout = LayoutStruct()
            layout.root = dict()
            layout.root['type'] = 'layout'
            layout.root['category'] = 'tab'
            layout.root['id'] = 1
            layout.root['items'] = floating_panels
            win = self.new_window_using_layout(layout)
            win.hide()            
            
        gui.qapp.deleteEmptyWindows(True)  

        self.panels.reselect_all()



from qtpy import QtCore, QtGui, QtWidgets

class yAxisLabel(QtWidgets.QGraphicsLineItem):
    
    def __init__(self, ticksX, ticksY, direction, parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        self.direction = direction
        
    def createGrid(self):
        self.grid = []
            
        pens = []
        pens.append(QtGui.QPen(QtGui.QColor(159,159,159), 0, QtCore.Qt.SolidLine))
        pens.append(QtGui.QPen(QtGui.QColor(191,191,191), 0, QtCore.Qt.DashLine))
        pens.append(QtGui.QPen(QtGui.QColor(223,223,223), 0, QtCore.Qt.DotLine))
        pens.append(QtGui.QPen(QtGui.QColor(159,159,159), 0, QtCore.Qt.SolidLine))
        pens.append(QtGui.QPen(QtGui.QColor(191,191,191), 0, QtCore.Qt.DashLine))
        pens.append(QtGui.QPen(QtGui.QColor(223,223,223), 0, QtCore.Qt.DotLine))        
        paths = []

        for ticklevel in range(3):
            paths.append(QtGui.QPainterPath())
            for i in self.ticksX[ticklevel][1]:
                if self.direction == 0:
                    paths[-1].moveTo(i, self.startY)
                    paths[-1].lineTo(i, self.stopY)               
                else:
                    paths[-1].moveTo(self.startY, i)
                    paths[-1].lineTo(self.stopY, i) 
                               
        for ticklevel in range(3):        
            paths.append(QtGui.QPainterPath())  
            for i in self.ticksY[ticklevel][1]:
                if self.direction == 0:
                    paths[-1].moveTo(self.startX, i)
                    paths[-1].lineTo(self.stopX, i)
                else:
                    paths[-1].moveTo(i, self.startX)
                    paths[-1].lineTo(i, self.stopX)
                
        for i in range(len(paths)):
            self.grid.append(QtWidgets.QGraphicsPathItem(paths[i]))
            self.grid[-1].setPen(pens[i])
            self.grid[-1].setZValue(-2)
            self.scene.addItem(self.grid[-1])



from qtpy import QtCore, QtGui, QtWidgets

from ...dialogs.formlayout import fedit
from ...panels.base import BasePanel
from ... import gui

class CmdHistTableModel(QtCore.QAbstractTableModel):
    def __init__(self):
        super().__init__()
        
    def rowCount(self, index):
        for row in gui.qapp.history.execfetch('SELECT COUNT() FROM CMDHIST'):
            count = row[0]
        return count
        
    def columnCount(self, index):
        return 2            
            
    def data(self, index, role):    
        if role == QtCore.Qt.DisplayRole:
            rownr, colnr = index.row(), index.column()
            for row in gui.qapp.history.execfetch(f'SELECT TIME, CMD FROM CMDHIST LIMIT {rownr}, 1'):
                value = row[colnr]
            return str(value)


class CmdHistTableView(QtWidgets.QTableView):
    def __init__(self):
        super().__init__()        
        
    def getSelectedCommands(self):
        selmodel = self.selectionModel()        
        lines = []
        for index in selmodel.selectedIndexes():
            text = self.model().data(index, QtCore.Qt.DisplayRole)
            lines.append(text)
            
        return '\n'.join(lines)            
        
    def mouseDoubleClickEvent(self, event):
        text = self.getSelectedCommands()
        console = gui.qapp.panels.selected('console')
        if not text.endswith('\n'):
            text += '\n'
        console.stdio.stdInputPanel.addText(text)
        
        
class CmdHistPanel(BasePanel):
    panelCategory = 'cmdhist'
    panelShortName = 'basic'
    userVisible = True     

    def __init__(self, parent, panid):
        super().__init__(parent, panid, type(self).panelCategory)
        
        self.initMenu()
        
        self.model = CmdHistTableModel()
        self.cmdhistview = CmdHistTableView()
        self.cmdhistview.setModel(self.model)
        self.setCentralWidget(self.cmdhistview)  
        
        self.statusBar().hide()
        
    def initMenu(self):        
        self.fileMenu = self.menuBar().addMenu("&File")
        
        self.addMenuItem(self.fileMenu, 'Close', self.close_panel,
            statusTip="Close this levels panel",
            icon = 'cross.png')         

        self.viewMenu = self.menuBar().addMenu("&View")      
        
        self.addMenuItem(self.viewMenu, 'Reset', self.reset)
        self.addMenuItem(self.viewMenu, 'Refresh', self.refresh)
        self.addMenuItem(self.viewMenu, 'Paste stdin', self.pasteStdin)
        
        self.addBaseMenu()        
        
    def reset(self):
        self.model.reset()        
                
    def refresh(self):
        self.cmdhistview.repaint()    
        
    def pasteStdin(self):
        text = self.cmdhistview.getSelectedCommands()            
        console = gui.qapp.panels.selected('console')
        console.stdio.stdInputPanel.setPlainText(text)




import threading
import sys, os
import ctypes
import logging
import textwrap
import pathlib
import mmap
import struct
import threading
import psutil

from multiprocessing.connection import Pipe
from multiprocessing.reduction import duplicate

from qtpy import QtGui, QtWidgets, QtCore
from qtpy.QtWidgets import QApplication, QShortcut, QDesktopWidget
from qtpy.QtGui import QIcon, QKeySequence
from qtpy.QtCore import Qt

from .. import gui, config, PROGNAME

from ..core.history import History
from ..core.watcher import CommandServer

from ..utils import new_id_using_keys
from ..utils.namedmutex import NamedMutex

from .qgc import QGarbageCollector
from .threadcom import HandOver
from .utils import getMenuAction, relax_menu_text, relax_menu_trace

from ..panels.window import MainWindow
from ..panels.panels import Panels
from ..panels.base import thisPanel

from ..dialogs.main import MainDialog

here = pathlib.Path(__file__).parent
respath = pathlib.Path(config['respath'])
logger = logging.getLogger(__name__)

        
class WaitCursorContext(object):
    def __init__(self, qapp, message=None):
        self.qapp = qapp
        self.message = message
        self.window = self.qapp.activeWindow()
        if not self.window in self.qapp.windows.values():
            self.window = None
        
    def __enter__(self):
        self.qapp.setOverrideCursor(QtCore.Qt.WaitCursor)
        if not self.message is None:
            #TO DO, bring this to the relevant window status bar
            #self.qapp.panels['console'][0].addText(f'{self.message}\n')
            logger.info(self.message)
            
            if not self.window is None:
                self.window.statusBar().showMessage(self.message)
            
        self.qapp.processEvents()
        
    def __exit__(self, exc_type, exc, exc_tb):
        self.qapp.restoreOverrideCursor()
        self.qapp.processEvents()
        
        if not self.window is None:
            self.window.statusBar().clearMessage()                       

      
class GuiApplication(QApplication):
    lastWindowId = 0

    def __init__(self, shell, argv):
        self.shell = shell
        
        super().__init__(argv)                
        
        self.windows = dict()
        self.panels = Panels(self)
        self.panelsDialog = MainDialog(self.panels)
        self.appIcon = QIcon(str(respath / 'logo' / 'logo_32px.png'))
        self.setWindowIcon(self.appIcon)
        self.handover = HandOver(self)
                
        self.history = History(self.shell.logdir.logpath)

        self.setFont(QtGui.QFont('MS Shell Dlg 2', pointSize=config['console']['fontsize']))

        if os.name == 'nt':
            # This is needed to display the app icon on the taskbar on Windows 7
            myappid = f'{PROGNAME}' # arbitrary string
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)    
            
        self.gc = QGarbageCollector(self)
        self.gc.enable()

        self.selected_group = dict()
        self.radio_group = dict()
        
        self.bindicon = QtGui.QIcon()
        self.bindicon.addFile(str(respath / 'icons' / 'bind_12px.png'), state=QIcon.On)
        self.bindicon.addFile(str(respath / 'icons' / 'broken_12px.png'), state=QIcon.Off)  
        
        self.resizeicon = QtGui.QIcon()
        self.resizeicon.addFile(str(respath / 'icons' / 'arrow_down_8px.png'), state=QIcon.Off)
        self.resizeicon.addFile(str(respath / 'icons' / 'arrow_right_8px.png'), state=QIcon.On)
        
        self.panelsDialog.showMinimized()
        self.focusChanged.connect(self.checkPanelActive)
        
        self.menuCallShortCuts = dict()

    def setShortCuts(self):        
        for layid in range(1,10):
            self.setShortCut(f"{config['shortcuts']['layout']['prefix']}{layid}",
                lambda layid=layid: self.panels.restore_state_from_config(layid))
            
        self.setPanelsDialogShortCut(config['shortcuts']['help'], ['Help', 'Help'])
        self.setPanelsDialogShortCut(config['shortcuts']['instance info'], ['Help', 'Instance Info'])
        self.setPanelsDialogShortCut(config['shortcuts']['panel preview'], ['Panel', 'Previews...'])
        self.setPanelsDialogShortCut(config['shortcuts']['window preview'], ['Window', 'Previews...'])
        
        self.setShortCut(config['shortcuts']['new panel'], lambda : self.showNewPanelDialog())
        
        for keySequence, menuCallParams in config["shortcutmenu"].items():
            for category, action_names in menuCallParams.items():
                if not category in self.menuCallShortCuts.keys():
                    self.menuCallShortCuts[category] = dict()
                self.menuCallShortCuts[category][relax_menu_trace(action_names)] = keySequence
                
            self.setShortCut(keySequence, lambda keySequence=keySequence: self.menuShortCutCall(keySequence))
        
    def setShortCut(self, keySequence, func):
        sc = QShortcut(QKeySequence(keySequence), self.panelsDialog, func)
        sc.setContext(Qt.ApplicationShortcut)
        
    def setPanelsDialogShortCut(self, keySequence, menuTrace):
        menuTrace = relax_menu_trace(menuTrace)
        
        def caller():
            action = getMenuAction(self.panelsDialog.menuBar(), menuTrace)
            action.trigger()
            
        action = getMenuAction(self.panelsDialog.menuBar(), menuTrace)
        action.setText(f'{action.text()}\t{keySequence}')
        self.setShortCut(keySequence, caller)
        
    def menuShortCutCall(self, keySequence):
        shortCutParams = config["shortcutmenu"][keySequence]

        if len(shortCutParams) == 1:
            category, action_names = next(iter(shortCutParams.items()))
            
        else:
            for category in reversed(self.panels.keys()):
                if category in shortCutParams.keys():
                    action_names = shortCutParams[category]
                    break
            else:
                category, action_names = None, []
                
        #logger.info(f'menuShortCutCall: {category} {action_names}')
                
        if category is None:
            return
            
        elif category == 'window':
            window = self.activeWindow()
            action = getMenuAction(window.windowMenu, action_names)
            action.trigger()
        else:
            action = self.panels.get_menu_action(category, None, action_names)
            if not action is None:
                action.trigger()
        
    def checkPanelActive(self, old, new):        
        #This function is probably more called then needed !
        if new is None: return
        try:
            panel = thisPanel(new)        
            if panel is None: return
            panel.select()
        except Exception as ex:
            #Do not print any error message in the gui
            #It would emit focusChanged and lead to infinite loop
            sys.__stdout__.write(str(ex))
            sys.__stdout__.flush()            
        
    @property 
    def mainWindow(self):
        return self.windows['main']
        
    def showDialog(self):
        self.panelsDialog.refresh()
        self.panelsDialog.exec_(self.activeWindow())    
        
    def showNewPanelDialog(self):
        self.panelsDialog.newMenu.exec_(QtGui.QCursor.pos())    

    def closePanel(self):
        window = self.activeWindow()
        panel = self.panels[window.activeCategory][window.activePanId]
        panel.close_panel()       
        
    def newWindow(self, name=None, parentName=None):
        if name is None:
            keys = [eval(k.split(' ')[1]) for k in self.windows.keys() if k.startswith('window ')]            
            key = new_id_using_keys(keys)
            name = f'window {key}'
        self.windows[name] = window = MainWindow(self, name, parentName)        
        return window  
        
    def getActiveWindow(self):
        for winname, window in self.windows.items():
            if window.isActiveWindow():
                return window           
                
    def cycleTagLevel(self):
        self.getActiveWindow().cycle_tag_level()                            
                
    def toggleFullScreen(self):
        self.getActiveWindow().fullScreen()            
        
    def toggleMenuStatusbar(self):
        self.getActiveWindow().toggleMenuStatusbar()            
        
    def toggleShortCuts(self):    
        self.panelsDialog.setVisible(not self.panelsDialog.isVisible())
        
    def deleteWindow(self, window):
        if isinstance(window, str):
            winname = window
        else:
            winname = window.name        
        window = self.windows.pop(winname)        
        container = self.panels.ezm.containers.pop(winname)
        window.deleteLater()
        
    def deleteEmptyWindows(self, check=False):
        for winname in list(self.windows.keys()):
            if winname == 'main':
                continue            
            window = self.windows[winname]
            
            if window.container.is_empty(check):                
                self.deleteWindow(winname)             
                
    def waitCursor(self, message=None):
        return WaitCursorContext(self, message)
                
    def hideWindow(self, window=None):
        if not window is None:
            window.hide()
        
        #Close if no window visible
        visibles = [win.isVisible() for win in self.windows.values()]
        
        if sum(visibles) == 0:
            self.panelsDialog.showNormal()                   
        

def eventloop(shell, init_code=None, init_file=None, console_id=0, pictures=None):
    """
    The GUI Process and Thread running the eventloop
    """            
    shell.logdir.find_log_path()
    
    qapp = GuiApplication(shell, sys.argv)           
    qapp.setShortCuts()
    qapp.newWindow('main')
            
    #To run in a new thread but on the same gui process
    #panid = qapp.mainWindow.newThread()
    qapp.mainWindow.show()
            
    desktopGeometry = QDesktopWidget().availableGeometry()
    qapp.mainWindow.resize(int(desktopGeometry.width()*3/5), int(desktopGeometry.height()*3/5))
    
    qtRectangle = qapp.mainWindow.frameGeometry()    
    centerPoint = desktopGeometry.center()
    qtRectangle.moveCenter(centerPoint)
    qapp.mainWindow.move(qtRectangle.topLeft())
    
    # Make sure the gui proxy for the main thread is created
    qapp.panels.restore_state_from_config('base')
       
    # if not config['debug']['skip_restore_perspective']:
        # if config['default_perspective'] != 'base':
            # qapp.panels.restore_state_from_config(config['default_perspective'])
    
    qapp.processEvents()    
    qapp.cmdserver = CommandServer(shell)
    
    if not init_file is None:
        cmd = {'cmd': 'execute_file', 'args': (init_file, console_id)}
        qapp.cmdserver.cmd_queue.put(cmd)
        
    if not init_code is None:
        cmd = {'cmd': 'execute_code', 'args': (init_code, console_id)}
        qapp.cmdserver.cmd_queue.put(cmd)        
        
    if not pictures is None:
        cmd = {'cmd': 'open_images', 'args': pictures}
        qapp.cmdserver.cmd_queue.put(cmd)            
        
    cmd = config.get('init_command')       
    qapp.cmdserver.cmd_queue.put(cmd)
    
    qapp.cmdserver.start_queue_loop(qapp)
    qapp.cmdserver.start(qapp)    
    exit_code = qapp.exec_()
    
    #Kill all the children    
    parent = psutil.Process(os.getpid())
    for child in parent.children(recursive=True):
        child.kill()
        
    from pylab import plt
    plt.close('all')
    
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
    print(f'Exiting {PROGNAME}. Releasing lock.')   
    shell.logdir.release_lock_file()    
    
            


from qtpy.QtCore import Qt
from qtpy.QtGui import QWindow
from qtpy.QtWidgets import QWidget
from qtpy import QtWebEngineWidgets

from ..base import BasePanel, CheckMenu

HTML_BANNER = '''<html>
<head>
<title>A Sample Page</title>
</head>
<body>
<h1>Hello, World!</h1>
<hr />
I have nothing to say.
</body>
</html>'''

class HtmlPanel(BasePanel):
    panelCategory = 'html'
    panelShortName = 'basic'

    def __init__(self, parent, panid):
        super().__init__(parent, panid, type(self).panelCategory)
        
        self.fileMenu = CheckMenu("File", self.menuBar())
        self.addMenuItem(self.fileMenu, "Close", self.close_panel,
            statusTip = "Close this html panel",
            icon = 'cross.png')
        
        self.webview = QtWebEngineWidgets.QWebEngineView()
        self.webview.setHtml(HTML_BANNER)
        self.setCentralWidget(self.webview)        
        self.addBaseMenu()


import sys
import threading
import logging
from queue import Queue

from qtpy.QtCore import QObject, QMetaObject, Slot, Qt
from qtpy.QtWidgets import QWidget

class ReturnLock:

    """
    Lock and return value object used to pass the return value on SignalCalls.
    """

    def __init__(self):
        self.lock = threading.Lock()
        self.lock.acquire()
        
    def waitOnReturn(self):
        self.lock.acquire()
        return self.value
        
    def releaseReturn(self, value):
        self.value = value
        self.lock.release()
        
    def locked(self):
        return self.lock.locked()

class HandOver(QObject):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.signal_call_queue = Queue()                   
        
    def send(self, block=True, func=None, *args, **kwargs):
        """
        :param block: wait on the func to return
        :param func: the reference to the function to be called by the eventloop
        :param *args: unnamed arguments for the function
        :param **kwargs: key word arguments for the function
        """
        returnlock = ReturnLock()
        self.signal_call_queue.put((returnlock, func, args, kwargs))         
        QMetaObject.invokeMethod(self, "receive", Qt.QueuedConnection)           
        if block:
            returnlock.waitOnReturn()            
            return returnlock.value
        else:
            return returnlock                
        
    @Slot()    
    def receive(self):
        """
        Receive returnlock, func, args, kwargs from signal_call_queue
        """
        (returnlock, func, args, kwargs) = self.signal_call_queue.get()
        returnvalue = None
        try:
            returnvalue = func(*args, **kwargs)
        except:
            logging.error(f'Error during gui call {func} {args} {kwargs}')
            raise
        finally:
            returnlock.releaseReturn(returnvalue)   


#-------------------------------------------------------------------------------
# Copyright 2021 Thomas Cools
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#-------------------------------------------------------------------------------

"""Bench Desk"""

from .version import VERSION_INFO
from .core.conf import config, configure
from .core.gui_proxy import gui
from .live import use, using

PROGNAME = 'Gamma Desk'
DOC_HTML = 'https://thocoo.github.io/gdesk-data/docs'
DOC_HTML_EXTRA = ['https://test.pypi.org/project/gamma-desk']

__release__ = "-".join(map(str, VERSION_INFO)).replace("-", ".", 2)
__version__ = ".".join(map(str, VERSION_INFO[:3]))

shell = None

def refer_shell_instance(shellinst):
    """refer_shell_instance"""
    global shell
    shell = shellinst



import os
import sys
import time
import collections
from pathlib import Path
import json
import importlib
import copy
import re
import logging

logger = logging.getLogger(__name__)

global config

config = dict()
config_objects = dict()
configured = False

here = Path(__file__).parent.absolute()

FIRST_CONFIG_FILE = here.parent / 'config' / 'defaults.json'

REQUIRED = [
    ('numpy', 'numpy'),
    ('matplotlib', 'matplotlib'),
    ('PySide2', 'PySide2'),
    ('qtpy', 'qtpy'),
    ('psutil', 'psutil'),
    ('numba', 'numba'),
    ('pyzmq', 'zmq'),
    ('gdesk', 'gdesk'),
]

PATHPATTERN = re.compile('path_\w*')

def deep_update(source, overrides):
    """
    Update a nested dictionary or similar mapping.
    Modify ``source`` in place.
    """
    for key, value in overrides.items():
        if isinstance(value, collections.Mapping) and value:
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            source[key] = overrides[key]
            if PATHPATTERN.match(str(key)) or key in ['respath']:
                if not source[key] is None:
                    source[key] = os.path.expandvars(source[key])
    return source

def deep_diff(dict1, dict2):
    """
    Difference of nested dictionary or similar mapping.
    """
    common_keys = set(dict1.keys()).intersection(set(dict2.keys()))
    dict2_keys_only = set(dict2.keys()).difference(set(dict1.keys()))

    result = dict((key, dict2[key]) for key in dict2_keys_only)

    for key in common_keys:
        value1 = dict1[key]
        value2 = dict2[key]
        if isinstance(value1, collections.Mapping):
            assert isinstance(value2, collections.Mapping)
            value = deep_diff(value1, value2)
            if len(value) > 0:
                result[key] = value
        elif value1 !=  value2:
            result[key] = value2

    return result

def list_packages(packages):
    for (package, base) in packages:
        found = importlib.util.find_spec(base)
        if found is None:
            logger.debug(f'{package}:\n    NOT FOUND')
        else:
            modified = os.path.getmtime(found.origin)
            modified_str = time.strftime('%Y-%b-%d %H:%M:%S',time.gmtime(modified))
            origin = Path(found.origin)
            install_dir = origin.parent.parent
            logger.debug(f'{package}:\n    {origin.parent}\n    {origin.name}: {modified_str}')
            for path in install_dir.glob(f'{package}*.dist-info'):
                logger.debug(f'    dist-info: {path.stem}')

def configure(**overwrites):
    global configured

    if configured:
        name = sys._getframe(1).f_globals['__name__']
        if not name in ['gdesk.core.tasks']:
            logging.warning(f'configure unexpected called from {name} but already configured, no reconfiguring done')
        return
    else:
        configured = True

    os.environ['GDESKROOT'] = str(here.parent.absolute())

    config_file = FIRST_CONFIG_FILE
    deep_update(config, load_config(config_file))

    prior_config_file = None
    next_config_file = overwrites.get('config_file', None) or config['next_config_file']

    while not next_config_file is None:
        config_file = Path(next_config_file).expanduser()
        if not config_file.exists():
            break
        if prior_config_file == config_file:
            logger.warn(f'config file {config_file} is loading itself, breaking it...')
            logger.warn(f'Set "next_config_file": null in {prior_config_file}')
            break
        logger.info(f'Loading config: {config_file}')
        deep_update(config, load_config(config_file))
        prior_config_file = config_file
        next_config_file = config['next_config_file']

    deep_update(config, overwrites)

    os.environ['QT_API'] = config['qt_api']
    logging.root.setLevel(config['logging_level'])

    if config['debug'].get('list_packages', False):
        list_packages(REQUIRED)

    #TO DO: import register_objects_in_ghawk2_init
    #from ghawk2.core.gui_proxy import register_objects_in_ghawk2_init
    #register_objects_in_ghawk2_init()

    # Importing matplotlib takes some time !
    # It also imports numpy
    import matplotlib
    matplotlib.use(config['matplotlib']['backend'])
    #This will also call the backend module
    #Which import ..panels.matplot
    import pylab

    #Configure and register plugins
    for panel_class_module in config["panel_class_modules"]:
        exec(f'import {panel_class_module}')

def save_config_json(path=None):
    current_config = copy.deepcopy(config)
    current_config['qapp'] = False
    current_config['next_config_file'] = None
    save_config = not_defaults(current_config)
    with open(path, 'w') as fp:
        json.dump(save_config, fp, indent=2)

def load_config(path):
    config_file = Path(path)

    if config_file.suffix in ['.cpy']:
        config_dict = eval(open(config_file).read())

    elif config_file.suffix in ['.json']:
        config_dict = load_config_json(config_file)

    return config_dict

def load_config_json(path=None):
    with open(path, 'r') as fp:
        loaded_config = json.load(fp)
    return loaded_config

def not_defaults(current_config):
    defaults = {}
    deep_update(defaults, load_config(FIRST_CONFIG_FILE))
    return deep_diff(defaults, current_config)



import numpy as np
from .fasthist import hist2d

stdquant = np.ndarray(13)
stdquant[0] = (0.0000316712418331200)  #-4 sdev
stdquant[1] = (0.0013498980316301000)  #-3 sdev
stdquant[2] = (0.0227501319481792000)  #-2 sdev
stdquant[3] = (0.05)
stdquant[4] = (0.1586552539314570000)  #-1 sdev or lsdev
stdquant[5] = (0.25)                   #first quartile
stdquant[6] = (0.50)                   #median
stdquant[7] = (0.75)                   #third quartile
stdquant[8] = (0.8413447460685430000)  #+1 sdev or usdev
stdquant[9] = (0.95)
stdquant[10] = (0.9772498680518210000) #+2 sdev
stdquant[11] = (0.9986501019683700000) #+3 sdev
stdquant[12] = (0.9999683287581670000) #+4 sdev
            

def get_standard_quantiles(arr, bins=64, step=None, quantiles=None):

    hist, starts, stepsize = hist2d(arr, bins, step, plot=False)              
    cumhist = np.cumsum(hist)       
    
    if quantiles is None:
        quantiles = stdquant  
    else:
        quantiles = np.array(quantiles)

    n = len(quantiles)
    npix = np.multiply.reduce(arr.shape)
    quantiles *= npix                                
    thresh = [0] * n
    
    #TO DO: speed up by using interpolation function of numpy
    for ind in range(n):
        thresh[ind] = starts[(cumhist < quantiles[ind]).sum()]
            
    return thresh
    
def get_sigma_range(arr, sigma=1, bins=64, step=None):    
    if sigma == 1:
        return get_standard_quantiles(arr, bins, step, (stdquant[4], stdquant[8])) 
    elif sigma == 2:
        return get_standard_quantiles(arr, bins, step, (stdquant[2], stdquant[10]))
    elif sigma == 3:
        return get_standard_quantiles(arr, bins, step, (stdquant[1], stdquant[11]))
    elif sigma == 4:
        return get_standard_quantiles(arr, bins, step, (stdquant[0], stdquant[12]))
        
        
def get_sigma_range_for_hist(starts, hist, sigma):
    cumhist = np.cumsum(hist)
    
    if sigma==1:
        quantiles = np.array((stdquant[4], stdquant[8]))
    elif sigma==2:
        quantiles = np.array((stdquant[2], stdquant[10]))
    elif sigma==3:
        quantiles = np.array((stdquant[1], stdquant[11]))
    elif sigma==4:
        quantiles = np.array((stdquant[0], stdquant[12]))               

    n = len(quantiles)
    npix = cumhist[-1]
    quantiles *= npix                                
    thresh = [0] * n
    
    #TO DO: speed up by using interpolation function of numpy
    for ind in range(n):
        thresh[ind] = starts[(cumhist < quantiles[ind]).sum()]
            
    return thresh    


# -*- coding: latin-1 -*-
#-------------------------------------------------------------------------------
# Name:        imageport.roi
# Purpose:     Region of Interest
#
# Author:      Thomas Cools
#
# Created:     01/08/2014
# Copyright:   (c) Thomas Cools 2014
# Licence:     <your licence>
#-------------------------------------------------------------------------------

from qtpy import QtCore, QtGui, QtWidgets

from ... import config

AUTO_APPLY = False

class SelRoiWidget(QtWidgets.QWidget):
    
    """
    Selection widget of a region of interest.
    """
    
    roiChanged = QtCore.Signal()
    roiRemoved = QtCore.Signal()
    
    def __init__(self, parent=None):
        #width and height are the dimensions of the image (not the roi)
        super().__init__(parent=parent)

        self.phase = 0
        self.solidColor = QtCore.Qt.white        

        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.newPhase)
        self.timer.setSingleShot(True)
        self.setMouseTracking(True)        
        #  is this timer also active when roi isn't visible ???

        self.initProps()
        self.initUI()
        self.hide()

    def initUI(self):
        self.scaleCursor = QtGui.QCursor(QtCore.Qt.SizeAllCursor)
        self.fillColor = QtGui.QColor(*config['roi color'])
        self.dashColor = QtGui.QColor(*config['roi color'])
        self.setAttribute(QtCore.Qt.WA_OpaquePaintEvent, False)
        
        self.get_context_menu = lambda: None        
        
    @property
    def vd(self):
        return self.parent().vd
        
    @property
    def selroi(self):
        return self.vd.selroi

    def initProps(self):
        self.dragSliceStartX = self.selroi.xr.start
        self.dragSliceStartY = self.selroi.yr.start
        self.dragStartX = 0
        self.dragStartY = 0
        self.mouseRightWasDown = False
        self.mouseMidWasDown = False
        self.mouseDoubleClicked = False

        self.overscan = 5
        self.createState = False

    def selectAll(self):
        self.selroi.reset()
        self.selroi.update_statistics()
        self.recalcGeometry()

    def newPhase(self):
        self.setAttribute(QtCore.Qt.WA_OpaquePaintEvent, True)
        
        self.phase = (self.phase + 1) % 8

        #self.parent().setUpdatesEnabled(False)
        #self.setUpdatesEnabled(True)                        
        
        #self.blockSignals(True)
        self.repaint(0,self.overscan,self.width(),1)
        self.repaint(self.overscan,0,1,self.height())
        self.repaint(0,self.height()-self.overscan-1,self.width(),1)
        self.repaint(self.width()-self.overscan-1,0,1,self.height())
        
        self.setAttribute(QtCore.Qt.WA_OpaquePaintEvent, False)
        #self.update()
        #self.parent().blockSignals(False)
        self.timer.start(100)
        #self.parent().setUpdatesEnabled(True)

    def setStartEndPoints(self, startX, startY, endX, endY):
    
        startX = int(round(startX))
        startY = int(round(startY))
        endX = int(round(endX))
        endY = int(round(endY))                

        if (startX > endX) and (startY > endY):
            pass

        else:
            if startY > endY:
                startY = endY
            if startX > endX:
                startX = endX

        self.selroi.xr.start = startX
        self.selroi.xr.stop = endX + 1
        self.selroi.yr.start = startY
        self.selroi.yr.stop = endY + 1

        self.clip()

    def clip(self):
        self.selroi.ensure_rising()
        self.selroi.clip()
        self.selroi.update_statistics()
        self.recalcGeometry()

    def asSliceTupple(self):
        return (slice(self.selroi.xr.start, self.selroi.xr.stop), \
            slice(self.selroi.yr.start, self.selroi.yr.stop))

    def recalcGeometry(self):
        x0 = round((self.selroi.xr.start - self.parent().dispOffsetX) * self.parent().zoomValue)
        y0 = round((self.selroi.yr.start - self.parent().dispOffsetY) * self.parent().zoomValue)
        x1 = round((self.selroi.xr.stop - self.parent().dispOffsetX) * self.parent().zoomValue)
        y1 = round((self.selroi.yr.stop - self.parent().dispOffsetY) * self.parent().zoomValue)
        width = max(abs(x1 - x0),1)
        height = max(abs(y1 - y0),1)

        self.setGeometry(min(x0,x1)-self.overscan, min(y0,y1)-self.overscan, width+2*self.overscan, height+2*self.overscan)

    def mousePressEvent(self, event):
        if (event.buttons() == QtCore.Qt.RightButton) and \
            not self.createState:
            #check if we are not doing setStartEndPoints
                self.createState = True

                self.edgePosition = self.checkNearEdge(event)
                self.setCursorShape(self.edgePosition)

                self.dragStartX  = event.globalX()
                self.dragStartY  = event.globalY()

                self.dragSliceStartX = self.selroi.xr.start
                self.dragSliceStartY = self.selroi.yr.start
                self.dragSliceEndX = self.selroi.xr.stop
                self.dragSliceEndY = self.selroi.yr.stop

                self.mouseRightWasDown = True
                self.repaint()
        elif (event.buttons() == QtCore.Qt.MidButton):
            self.dragStartX  = event.globalX()
            self.dragStartY  = event.globalY()
            self.mouseMidWasDown = True
            #propagated up the parent widget
            event.ignore()
        else:
            #propagated up the parent widget
             event.ignore()

    def checkNearEdge(self, event):
        x = event.pos().x()
        y = event.pos().y()
        (x0, y0, x1, y1) = self.getRelativeCoord()

        if abs(x0 - x1) <= 10:
            hori = 1
        elif abs(x0 - x) < 5:
            hori = 0
        elif abs(x1 - x) < 5:
            hori = 2
        else:
            hori = 1

        if abs(y0 - y1) <= 10:
            vert = 1
        elif abs(y0 - y) < 5:
            vert = 0
        elif abs(y1 - y) < 5:
            vert = 2
        else:
            vert = 1

        return vert * 3 + hori

    def setCursorShape(self, edgePosition):
        if edgePosition in (0, 8):
            self.setCursor(QtCore.Qt.SizeFDiagCursor)
        elif edgePosition in (1, 7):
            self.setCursor(QtCore.Qt.SizeVerCursor)
        elif edgePosition in (2, 6):
            self.setCursor(QtCore.Qt.SizeBDiagCursor)
        elif edgePosition in (3, 5):
            self.setCursor(QtCore.Qt.SizeHorCursor)
        elif edgePosition == 4:
            self.setCursor(QtCore.Qt.SizeAllCursor)

    def getMouseShifts(self, event, manhattan=False):
        if manhattan:
            if abs(self.dragStartX - event.globalX()) > abs(self.dragStartY  - event.globalY()):
                self.dragEndX = event.globalX()
                self.dragEndY = self.dragStartY
            else:
                self.dragEndX = self.dragStartX
                self.dragEndY = event.globalY()
        else:
            self.dragEndX = event.globalX()
            self.dragEndY = event.globalY()
        shiftX = round((self.dragEndX - self.dragStartX) / self.parent().zoomValue)
        shiftY = round((self.dragEndY - self.dragStartY) / self.parent().zoomValue)
        return (shiftX, shiftY)

    def mouseMoveEvent(self, event):
        if event.buttons() == QtCore.Qt.RightButton:            
            if event.modifiers() & QtCore.Qt.ShiftModifier:
                shiftX, shiftY = self.getMouseShifts(event, manhattan=True)
            else:
                shiftX, shiftY = self.getMouseShifts(event, manhattan=False)
                
            if self.edgePosition == 0:
                self.setStartEndPoints(self.dragSliceStartX  + shiftX, self.dragSliceStartY + shiftY,\
                    self.dragSliceEndX -1, self.dragSliceEndY -1)
            elif self.edgePosition == 1:
                self.setStartEndPoints(self.dragSliceStartX, self.dragSliceStartY + shiftY,\
                    self.dragSliceEndX -1, self.dragSliceEndY -1)
            elif self.edgePosition == 2:
                self.setStartEndPoints(self.dragSliceStartX, self.dragSliceStartY + shiftY,\
                    self.dragSliceEndX + shiftX - 1, self.dragSliceEndY -1)
            elif self.edgePosition == 3:
                self.setStartEndPoints(self.dragSliceStartX  + shiftX, self.dragSliceStartY,\
                    self.dragSliceEndX - 1, self.dragSliceEndY -1)
            elif self.edgePosition == 4:
                #moving
                #limiting the shift to the borders of the image
                if (self.dragSliceStartX  + shiftX) < 0:
                    shiftX = - self.dragSliceStartX
                if (self.dragSliceEndX + shiftX - 1) >= self.selroi.xr.maxstop:
                    shiftX = self.selroi.xr.maxstop - self.dragSliceEndX
                if (self.dragSliceStartY  + shiftY) < 0:
                    shiftY = - self.dragSliceStartY
                if (self.dragSliceEndY + shiftY - 1) >= self.selroi.yr.maxstop:
                    shiftY = self.selroi.yr.maxstop - self.dragSliceEndY                        
                self.setStartEndPoints(self.dragSliceStartX  + shiftX, self.dragSliceStartY + shiftY,\
                    self.dragSliceEndX + shiftX - 1, self.dragSliceEndY + shiftY - 1)
            elif self.edgePosition == 5:
                self.setStartEndPoints(self.dragSliceStartX, self.dragSliceStartY,\
                    self.dragSliceEndX + shiftX - 1, self.dragSliceEndY  -1)
            elif self.edgePosition == 6:
                self.setStartEndPoints(self.dragSliceStartX  + shiftX, self.dragSliceStartY,\
                    self.dragSliceEndX - 1, self.dragSliceEndY + shiftY - 1)
            elif self.edgePosition == 7:
                self.setStartEndPoints(self.dragSliceStartX, self.dragSliceStartY,\
                    self.dragSliceEndX - 1, self.dragSliceEndY + shiftY - 1)
            elif self.edgePosition == 8:
                self.setStartEndPoints(self.dragSliceStartX, self.dragSliceStartY,\
                    self.dragSliceEndX + shiftX - 1, self.dragSliceEndY + shiftY - 1)
            self.repaint()
        else:
            event.ignore()

    def mouseReleaseEvent(self, event):
        self.createState = False
        
        contextMenu = self.get_context_menu()

        if self.mouseRightWasDown:
            self.mouseRightWasDown = False
            shiftX, shiftY = self.getMouseShifts(event)
            if (shiftX == 0) and (shiftY == 0):   
                pos = QtGui.QCursor.pos()
                pos.setX(pos.x() - 10)
                pos.setY(pos.y() - 10)
                
                if not contextMenu is None:
                    contextMenu.exec_(pos)              
                    
                #return None
                #self.parent().myContextMenuEvent(event)
            else:
                self.clip()
                if AUTO_APPLY:
                    self.applyRoi()            
                self.roiChanged.emit()

        if self.mouseMidWasDown:
            self.mouseMidWasDown = False
            shiftX, shiftY = self.getMouseShifts(event)
            if (shiftX == 0) and (shiftY == 0):
                self.parent().zoomAuto()
        
        self.unsetCursor()
        self.repaint()
        event.ignore()
        
    def release_creation(self):
        self.createState = False        
        self.clip()
        if AUTO_APPLY:
            self.applyRoi()
        self.roiChanged.emit()
        self.repaint()                
        
    def hideRoi(self):
        self.selroi.reset()
        if AUTO_APPLY:
            self.applyRoi()
        self.hide()
        self.unsetCursor()
        self.roiRemoved.emit()
        self.repaint()   

    def applyRoi(self):
        self.parent().vd.applyroi()

    def paintEvent(self, e):
        qp = QtGui.QPainter()
        qp.begin(self)
        self.drawRoi(qp)
        qp.end()        

    def getRelativeCoord(self):
        x0 = self.overscan
        y0 = self.overscan
        x1 = self.size().width()-1-self.overscan
        y1 = self.size().height()-1-self.overscan
        return (x0, y0, x1, y1)

    def drawRoi(self, qp):
        pensolid = QtGui.QPen(self.solidColor, 1, QtCore.Qt.SolidLine)
        pendash = QtGui.QPen(self.dashColor, 1, QtCore.Qt.CustomDashLine)
##        pendash.setDashPattern([3, 3, 1, 5, 3, 3, 5, 1])
        pendash.setDashPattern([4,4])

        x0 = self.overscan
        y0 = self.overscan        
        #max-> keep is visible even if it is smaller then 1x1
        x1 = max(self.size().width() - 1 - self.overscan, x0 + 1)
        y1 = max(self.size().height() - 1 - self.overscan, y0 + 1)

        pendash.setDashOffset(8-self.phase)

        polygonRect = QtGui.QPolygon()
        polygonNe = QtGui.QPolygon()
        polygonSw = QtGui.QPolygon()

        polygonRect << QtCore.QPoint(x0, y0) << QtCore.QPoint(x1, y0)\
            << QtCore.QPoint(x1, y1) << QtCore.QPoint(x0, y1)
        polygonNe << QtCore.QPoint(x0, y0) << QtCore.QPoint(x1, y0)\
            << QtCore.QPoint(x1, y1)
        polygonSw << QtCore.QPoint(x0, y0)  << QtCore.QPoint(x0, y1)\
            << QtCore.QPoint(x1, y1)


        if self.createState:
            self.setAttribute(QtCore.Qt.WA_OpaquePaintEvent, False)
            qp.setOpacity(0.25)
            qp.fillRect(x0, y0, x1-x0+1, y1-y0+1, self.fillColor)
            qp.setOpacity(0.5)
        else:
            qp.setOpacity(1.0)
            self.timer.start(100)
            

        qp.setPen(pensolid)
        qp.drawPolygon(polygonRect)
        qp.setPen(pendash)
        qp.drawPolyline(polygonNe)
        qp.drawPolyline(polygonSw)
        
        #self.parent().blockSignals(False)



import sqlite3
import time
import shutil
from pathlib import Path

try:
    import msvcrt

except ModuleNotFoundError:
    #Not on Linux
    pass

from .conf import config

class LogDir(object):
    def __init__(self, rootpath):
        self.rootpath = Path(rootpath)
        if not self.rootpath.exists():
            self.rootpath.mkdir(parents=True)

    def make_lock_file(self, lock_file):
        """
        Make a empty lock file in the logpath.

        A windows readlock is attached to the file.
        The lock will be removed by windows if the process dies.
        When booting another instance of Ghawk, this lock file will
        be used (the possibility to removed it) to detect if the
        logpath is currently in use by another Ghawk process.
        """
        self.lock_file = lock_file
        self.lock_file.touch()
        self.lock_file_ptr = open(str(self.lock_file), 'r')
        msvcrt.locking(self.lock_file_ptr.fileno(), msvcrt.LK_RLCK, 1024)
        
    def release_lock_file(self):
        """
        Release and remove the current lock file.

        When this process dies, the lock is removed by windows but the
        file is not removed.
        """    
        msvcrt.locking(self.lock_file_ptr.fileno(), msvcrt.LK_UNLCK, 1024)
        self.lock_file_ptr.close()
        self.lock_file.unlink()
        
    def get_active_lock_files(self):
        active_locks = []
        for p in self.rootpath.glob('log.*'):
           if (p / 'cmdlog.lock').exists():
              try:
                  (p / 'cmdlog.lock').unlink()
              except PermissionError:
                  active_locks.append(p / 'cmdlog.lock')
        return active_locks
        
    def remove_inactive_log_dirs(self):
        active_locks = self.get_active_lock_files()
        for p in self.rootpath.glob('log.*'):
            if not (p / 'cmdlog.lock') in active_locks:
                shutil.rmtree(str(p))
            
    def find_log_path(self):
        """
        Find a suitable rootpath/log.N path.

        Which logpath contains an active lock file?
        Create a new log.N if needed.
        """
        priorlogpath = None
        for n in range(config['max_log_paths']):
            logpath_propose = self.rootpath / ('log.%d' % n)
            if logpath_propose.exists():
                # Is the path in use by other process?
                if not (logpath_propose / 'cmdlog.lock').exists():
                    logpath = logpath_propose
                    break
                else:
                    try:
                        (logpath_propose / 'cmdlog.lock').unlink()
                        # succesfull delete of lock file
                        # no process was using it
                        logpath = logpath_propose
                        break
                    except:
                        # Logs from this logpath should be copied to own logs.
                        priorlogpath = logpath_propose
            else:
                logpath_propose.mkdir()
                logpath = logpath_propose
                break
                
        else:
            print(f'Maximum number of {config["max_log_paths"]} log dirs reached')
            print(f'Please, remove deprecated dirs')
            print(f'{self.rootpath }')
            raise SystemExit()
                
        self.make_lock_file(logpath / 'cmdlog.lock')        
        self.logpath = logpath
        self.priorlogpath = priorlogpath
                
        return logpath, priorlogpath

class History(object):
    def __init__(self, logdir):
        self.init_server(logdir)
        
    def init_server(self, logdir=None):
        if logdir is None:
            self.server_file = None
            self.server = sqlite3.connect(':memory:')            
            self.define_tables()
        else:
            logdir = Path(logdir)
            if not logdir.exists():
                logdir.mkdir(parents=True)                
            logdir = logdir.absolute()
            self.server_file = logdir / 'ghhist.db'
            #Python 3.6 doesn't understand Path
            self.server = sqlite3.connect(str(self.server_file))
            self.define_tables()
            
    def import_command_history(self, other_logfile):  
        self.server.execute('ATTACH [%s] as OTHERDB' % str(other_logfile))
        q = 'INSERT INTO CMDHIST (TIME, CMD) SELECT TIME, CMD FROM [OTHERDB].[CMDHIST]'
        self.server.execute(q)
        self.server.commit()
        self.server.execute('DETACH OTHERDB')            
            
    def define_tables(self):
        query = """CREATE TABLE IF NOT EXISTS CMDHIST (
ID INTEGER PRIMARY KEY, TIME TEXT, CMD TEXT)"""
        self.server.execute(query)
        
        query = """CREATE TABLE IF NOT EXISTS PATHHIST (
ID INTEGER PRIMARY KEY, CATEGORY TEXT, TIME TEXT, PATH TEXT)"""
        self.server.execute(query)
        
        query = """ALTER TABLE PATHHIST ADD COLUMN CATEGORY TEXT"""
        try:
            self.server.execute(query)
        except:
            pass        
            
    def execfetch(self, query, parameters=()):        
        cur = self.server.cursor()
        cur.execute(query, parameters)
        row = cur.fetchone()
        while row !=  None:
            yield row
            row = cur.fetchone()        

    def logcmd(self, cmd):
        now = time.strftime('%Y-%m-%d %H:%M:%S')
        query = "INSERT INTO CMDHIST (TIME, CMD) VALUES (?, ?)"
        self.server.execute(query, (now, cmd,))
        self.server.commit()
        self.skip = -1
        rowid = None
        for col in self.execfetch("SELECT last_insert_rowid()"):
            rowid = col[0]
        return rowid
         
    def storepath(self, path, delete_old_entry=True, category='image'):
        if delete_old_entry:
            self.server.execute("DELETE FROM PATHHIST WHERE PATH = ?", (path,))        
            
        now = time.strftime('%Y-%m-%d %H:%M:%S')
        query = "INSERT INTO PATHHIST (CATEGORY, TIME, PATH) VALUES (?, ?, ?)\n"
        self.server.execute(query, (category, now, path,))
        self.server.commit()

        rowid = None
        for col in self.execfetch("SELECT last_insert_rowid()"):
            rowid = col[0]
        return rowid      

    def yield_recent_paths(self, count=20, category='image'):        
        for row in self.execfetch("SELECT ID, TIME, PATH FROM PATHHIST WHERE CATEGORY = ? ORDER BY ID DESC LIMIT ?", (category, count)):
            yield row
            
        
    def retrievecmd(self, part='', from_id=None, distinct=True, back=True, prefix=True):        
            
        query = self.make_retrieve_query(1, part, from_id, distinct, back, prefix)
        
        for cmdid, cmd in self.execfetch(query):
            return cmdid, cmd
            
        return from_id, part

    def tail(self, count=20, part='', from_id=None, distinct=False, back=True, prefix=True, reverse=True):
        cmds = []
        
        query = self.make_retrieve_query(count, part, from_id, distinct, back, prefix)
            
        for row in self.execfetch(query):
            cmds.append(row)
            
        if reverse:
            return cmds[::-1]
        else:
            return cmds
            
    def make_retrieve_query(self, count=20, part='', from_id=None, distinct=False, back=True, prefix=True):
    
        if back:
            order = 'DESC'
            if not (from_id is None or from_id == 0):
                rng = f" WHERE ID < {from_id}"
            else:
                rng = ''
        else:
            order = 'ASC'
            if not (from_id is None or from_id == 0):
                rng = f" WHERE ID > {from_id}"            
            else:
                rng = ''
                
        if prefix:
            wild = ''
        else:
            wild = '%'
            
        part = part.replace("'","''")
            
        if distinct:
            query = f"SELECT LASTID AS ID, CMD FROM (SELECT MAX(ID) AS LASTID, CMD FROM CMDHIST WHERE CMD LIKE '{wild}{part}%' GROUP BY CMD ORDER BY LASTID {order}){rng} LIMIT {count}"
        else:
            query = f"SELECT ID, CMD FROM (SELECT ID, CMD FROM CMDHIST WHERE CMD LIKE '{part}%'{rng} ORDER BY ID {order}){rng} LIMIT {count}"

        return query
            